#ifndef ASTROMETRY_DEFS
#define ASTROMETRY_DEFS



#include "OmSourceParameters.h"


#include "Dal.h"
#include <f90cxx.h>


class Astrometry {
public:
Astrometry() : exposure(0.0), xShift(0.0), xShiftErr(0.0), 
                yShift(0.0), yShiftErr(0.0), 
	        rmsResidual(0.0),
                nMatches(0), nStars(0), okay(false) {}; 
                     
    float exposure; 
    float xShift;
    float xShiftErr;
    

    float yShift;
    float yShiftErr;
    float rmsResidual;
    int nMatches;
    int nStars;
    string filter;
    string obsId;
    bool okay;
};






/*****************************************************
*
* Computes the equatorial rectangular coordinates
*
*****************************************************/
class EquatorialCoordinates
{
public:
    double ra;
    double dec;
    double x;       
    double y;
    double z;
    EquatorialCoordinates() : ra(0.0), dec(0.0), x(0.0), y(0.0), z(0.0) {};

    EquatorialCoordinates(double raIn, double decIn);

    EquatorialCoordinates(double xIn, double yIn, double decIn);
    EquatorialCoordinates(const EquatorialCoordinates &);






    ~EquatorialCoordinates(){};

     
    void computeEquatorialCoordinates();
    
    /**************************************************************
    *
    **************************************************************/    
    void computeSphericalCoordinates();


    double arcLength( EquatorialCoordinates & a);


};    

class data : public EquatorialCoordinates {
public:

    data() : EquatorialCoordinates(0.0, 0.0), index(0), error(0.0), significance(0.0), extension(0), obsId("") {};  
    data( int n, double a, double b, float e = 0.0, float s = 0.0, Integer1 k = 0, string obs = "") : 
      EquatorialCoordinates(a, b), index(n), error(e), significance(s), extension(k), obsId(obs)  { };

    int index;    
    float error, significance;
    Integer1 extension;
    string obsId;
};






class matchedData {
public:

    matchedData() : index(0), distance(0.0)  {};

    
    matchedData( int ind, double r) : 
      index(ind), distance(r)  {};
    int index;
    double distance;
 

};

bool compare(const matchedData &a, const matchedData & b);

bool compare1(const data &a, const data & b);

bool compare2(const OmSourceParameter &a, const OmSourceParameter & b);





#endif



