/*****************************************************************************
* matchlists.cc - Given 2 source lists finds all matched
*                 within a given tolerance in arc secs
* 
*
******************************************************************************/


#include "OmSourceParameters.h"
#include <iostream>
#include <iomanip>
#include "param.h"
#include <vector>
#include <map>
#include <cmath>
#include<fstream>


#include "utility.h"
#include "astrometrydefs.h"
#include "matchlists.h"
#include "statistics.h"

/**********************************************
*
* int Locate                                 
*
**********************************************/


unsigned locate(vector<double> &xx, unsigned n, double x, unsigned jl, unsigned & ju)
{



  //  static int count = 0;

    if (n == 1) return 0;
    if (xx[0] > x) return 0;
    if (xx[n - 1] < x) return n - 1;

    unsigned jm;



    ju = n - 1;
    
    while (ju - jl > 1) {
      jm = (ju + jl) >> 1;
      if (xx[jm] < x)
          jl = jm;
      else
          ju = jm;
    }

   
    //    if(count < 100)
    //    {
    //      count++;
    //      cout << "x=" << x << " jl=" << jl << " ju=" << ju << " n=" << n << endl;

    //    } 
      



    return jl;


};
/*********************************************************************
*
* Matches two lists
* tolerance is the maximum allowed separation between 2 sources (arcsecs)
*
*
**********************************************************************/
void MatchLists::matchLists(double tolerance) 
{

  //  cout << "Matchlists- tolerance=" << tolerance << endl; 

       
    _matchedSource.resize( _listData.size() );
 

    /*****************************************************
    *
    * Loop through all the sources in the source-list
    *
    ****************************************************/
    vector<double> catRa;

    for(unsigned i = 0; i < _catData.size(); i++)
    {
        catRa.push_back( _catData[i].ra );  
    
    }


    //    unsigned jl = 0;

    //    unsigned ju = catRa.size();
    double t1 = 2.0 * tolerance * degreesToRadians / 3600.0;
 
 
 
    //    int n = listData.size();

  

   
    /*********************************************************************************
    * The following will supply a list of the matched sources that have a match with 
    * a particualr catalogue source
    **********************************************************************************/
    map<unsigned, vector<unsigned> > catMatched;


  

    /*************************************************************
    *
    * Loop through all the OM catalogue sources
    *
    *************************************************************/


    int nMatched = 0;

  
    unsigned jl = 0;
    unsigned ju = _catData.size() - 1;
   
   

    for(unsigned i = 0; i < _listData.size(); i++)
    {

     
        /*****************************************************
	*
        ******************************************************/
      // double t2 = t1 / max(0.01, cos( _listData[i].dec ) );
	//      unsigned ju = _catData.size() - 1;
        
	//	if( _listData[i].ra >= t2 && (_listData[i].ra + t2) <= TWO_PI )
	//	{

 
	//    jl = locate(catRa, (unsigned)catRa.size(), _listData[i].ra - t2, jl, ju);
          
        //     ju = locate(catRa, (unsigned)catRa.size(), _listData[i].ra + t2, jl, ju);
          
	    
	//	}
	//	else
	//	{
	//   jl = 0;
            
	//	}

            
	//	jl = 0;
        //ju = _catData.size() - 1;

	/*************************************************************
	*
	* Loop through all the catalogue sources
        *
        *************************************************************/
        

        for(unsigned j = jl; j <= ju; j++)
	{
	  
	    if( abs( _listData[i].dec - _catData[j].dec ) > t1) continue; 
	  	  
	    double d = _listData[i].arcLength( _catData[j] );

	    
            if(d <= tolerance)  
	    {
	      
                  
		/********************************************************************
	        * Remember that source i has been matched with catalogue source j
                *
                ******************************************************************/
		_matchedSource[i].push_back( matchedData(j, d) ); 
         
		//       catMatched[j].push_back( i ); // Remember that catalogue source j has been matched with matched source i               
		//		cout << "Line 191 Matched source " << i << " with cat source " << j << " distance=" << d << endl;


            }

        }     // End of looping through the catalogue sources

        if( _matchedSource[i].size() > 0) {
       
      
            nMatched++;
	}
        else
	{
	  //  cout << "Not matched!\n";
	  // exit(0); 

        }
      

        if( (i+1) % 1000 == 0)
	{ 
	   errstr << "Number matched=" << nMatched
	          << Sas::message(Msg::App, Msg::Sparse);


        }
    } // The end of matching all the sources
        

    errstr << "Total number of sources in SRCLIST1 matched=" << nMatched  
	   << Sas::message(Msg::App, Msg::Sparse);

 

    /************************************************************************
    *
    *
    * Sort the matched sources for each source in SRCLIST1 in order of increasing radius
    *
    * 
    **************************************************************************/    
    for (unsigned i = 0; i < _matchedSource.size(); i++)
    {

        std::sort( _matchedSource[i].begin(), _matchedSource[i].end(), compare); 
 
    }
   
   
    /*********************************************************************
    *
    * Some sources may have more than one match
    * Need to refine the list
    *
    *********************************************************************/ 
    for (unsigned i = 0; i < _matchedSource.size(); i++) // Loop through the list of matches
    {
	



        if(_matchedSource[i].size() == 0) continue;   // No matches to this source
        vector< matchedData > tempSource = _matchedSource[i];

	//	cout << i << ") matchedsource[0].index=" << _matchedSource[i][0].index << endl;
      
    

        /*****************************************************************
	*
	* Check each counterpart to see if it is also a counterpart of 
        * another source
        * Go through in reverse order 
        ******************************************************************/
         
        for (int j = (int)_matchedSource[i].size() - 1; j >= 0; j--)
	{
                
	   
	    /***************************************************************************
	    *
	    * Loop through all the other sources that may have a match with source j
            *
            ****************************************************************************/	 
	    int k1 = _matchedSource[i][j].index;              // The catalogue number

            if( catMatched[k1].size() == 0) continue;        // Only one match with this catalogue source

            double r1 = _matchedSource[i][j].distance;        
     
	    //	    cout << "Line 343 j=" << j << " k1=" << k1 << " r1=" << r1 << endl;    	  
                 
            for(unsigned k = 0; k < catMatched[k1].size(); k++)	  
            {

                if(i == k) continue;

		          
                vector< matchedData > tempCat = _matchedSource[k];

		  
		/************************************************************************************
		*
		* Loop through all the counterparts of source k to see if there is a counterpart
                *
                *************************************************************************************/  
                for (unsigned l = 0; l < _matchedSource[k].size(); l++)
	        {
	           
		      
		    if( k1 == _matchedSource[k][l].index)
		    {
		        if( r1 < _matchedSource[k][l].distance) 
			{

                      
		    
			      //		      cout << "Line 393\n";
                            tempCat.erase( tempCat.begin() + l);
                            break;
                        }
                        else
			{
			      //   cout << "Line 399\n";
                         
    
			    tempSource.erase(tempSource.begin() + j); //  We have removed this counterpart so 
			                                               // can go on to the next counterpart of source i
                            goto finished; 
			} 

                    } 

                } 


		    
		_matchedSource[k] = tempCat;

		//		cout << "Line 391 i=" << i << " k=" << k << endl;		    
              
            }
	   
	finished:;           
        } // End of looping through all the counterparts of source i 
        
     
    }

    //    cout << "matchedsource.size()=" << _matchedSource.size() << endl;

}; 
/********************************************************************************
*
* Gets the matches for a the given tolerance
*
*
*********************************************************************************/ 
int MatchLists::getMatches(double tolerance, vector<data> &matchedListData, vector<data> &matchedCatData)
{

    matchedListData.clear();
    matchedCatData.clear();

    int nMatches = 0;

    for(unsigned i = 0; i < _matchedSource.size(); i++)
    {     
        if( _matchedSource[i].size() > 0) // Each source in SRCLIST1 can have one or more sources from SRCLIST2 matched with it
    	{
       
	    if( _matchedSource[i][0].distance <= tolerance)
	    { 
                matchedListData.push_back( data( i, _listData[i].ra, _listData[i].dec, _listData[i].error, _listData[i].significance, _listData[i].extension) );
                int k = _matchedSource[i][0].index;

                matchedCatData.push_back( data( k, _catData[k].ra, _catData[k].dec, 0.0, 0.0, (int8)0) );
                nMatches++;
 
		//		cout << i << ") nmatches=" << nMatches << " k=" << k << " ralist1=" << _listData[i].ra * radiansToDegrees 
		//         << " dec=" << _listData[i].dec * radiansToDegrees 
		//      << " catData.ra=" << _catData[k].ra * radiansToDegrees << " dec=" << _catData[k].dec * radiansToDegrees << endl;
           }
    

        }

    }
    /************************************************************
    *
    * Compute the mean offsets between the matched coordinates
    *
    ************************************************************/

    computeOffsets(matchedListData, matchedCatData);

    return nMatches;


}
/***********************************************************************************
*
*
* Compute the offsets between the matched list and catalogue data
*
*
*************************************************************************************/
void MatchLists::computeOffsets( vector<data> &matchedListData, vector<data> &matchedCatData)
{
    _raMeanOffset = 0.0;
    _decMeanOffset = 0.0;

    if(matchedListData.size() == 0) return;

    vector<double> data(matchedListData.size() );

    double _mean = 0.0;
    for(unsigned i = 0; i < matchedListData.size(); i++)
    {

        double raDiff = matchedListData[i].ra - matchedCatData[i].ra;
        _raMeanOffset += raDiff;
        double decDiff = matchedListData[i].dec - matchedCatData[i].dec;
        _decMeanOffset += decDiff;
      
        double d = distance( matchedListData[i].ra, matchedListData[i].dec, 
                             matchedCatData[i].ra, matchedCatData[i].dec);
        data[i] = d;  
        _mean += d;
 
    }

    _mean /=  double(matchedListData.size() );
    _median = Median( data );
    _point05Percentile = Percentile( data, 0.05);
    _point95Percentile = Percentile( data, 0.95);

  
    _raMeanOffset  *= (radiansToArcSecs / double(matchedListData.size() ) );
    _decMeanOffset *= (radiansToArcSecs / double(matchedListData.size() ) );


    errstr << "Mean RA offset (list1-list2)=" << _raMeanOffset / cos(matchedListData[0].ra)  << " (arcsecs) "
           << "Mean DEC offset (list1-list2)=" << _decMeanOffset << " (arcsecs)\n "
	   << Sas::message(Msg::App, Msg::Sparse);

    errstr << "Mean separation=" << _mean << " median=" << _median 
	   << " 0.05 percentile=" << _point05Percentile << " 0.95=" 
           << _point95Percentile << " (arcsecs)" << endl; 
}

/***********************************************************************************
*
*
* Removes the offsets from the catalogue data
*
*
*************************************************************************************/
void MatchLists::removeOffsets()
{



  //    cout << "Remove offsets ra offset=" << _raMeanOffset << " dec=" << _decMeanOffset << endl;


    double dx = degreesToRadians * _raMeanOffset / 3600.0;
    double dy = degreesToRadians * _decMeanOffset / 3600.0;


    for(unsigned i = 0; i < _catData.size(); i++)
    {
        _catData[i].ra  += dx;
        _catData[i].dec += dy;

_catData[i].computeEquatorialCoordinates();




    }
}
/********************************************************************************
*
*
*
*********************************************************************************/ 
int MatchLists::getAllMatches(double tolerance, 
                              vector<data> &matchedListData, 
                              vector<vector<data> > &matchedCatData)
{

    matchedListData.resize(0);

    matchedCatData.resize(0);


    int nMatches = 0;
  


 

    
    for(unsigned i = 0; i < _matchedSource.size(); i++)
    {

    
        if( _matchedSource[i].size() > 0)
    	{
       
            if( _matchedSource[i][0].distance <= tolerance)
	    {
	        
	        matchedCatData.resize( nMatches + 1);  
                matchedListData.push_back( data( i, _listData[i].ra, _listData[i].dec, _listData[i].error, _listData[i].significance, _listData[i].extension) );
               
    


	        for(unsigned j = 0; j < _matchedSource[i].size(); j++)
	        {
	            if( _matchedSource[i][j].distance <= tolerance)
		    {
                        int k = _matchedSource[i][0].index;

                        matchedCatData[nMatches].push_back( data( k, _catData[k].ra, _catData[k].dec, 0.0, 0.0, (int8)0) );
                       
        
                    }
    
                }
                nMatches++;
	     
            }
        }

    }



    //    cout << "Matchlists.cc line 552 nmatches=" << nMatches << endl;
   
    return nMatches;


}




