/*************************************************************************
*
* Created 2008 July 25th by CB
*
*************************************************************************/
#include "ommergelists.h"
#include <iostream>
#include <vector>
#include <map>
#include <fstream>
#include <sstream>
#include <cstdio>
#include <iomanip>


/********************************************************
*
*
*
*
*
********************************************************/
#include "Dal.h"
#include "sourcelist.h"     // For various routines for FITS input/output using Dal interface 
#include "utility.h"        // For routines such as median, mean, etc
#include "astrometrydefs.h" // Some definitions for source-matching 
#include "matchlists.h"     // For matching 2 source-lists using RAs
#include "sorttable.h"      // For sorting SRCLIST table
#include "graph.h"          // For pgplot graphics
#include "omfilter.h"       // For OM filters
//#include "skewness.h"


void checkTable(Table *table, string colName);



/****************************************************
*
* The following file contains all the main function prototypes
* for the functions used in this file 
*
****************************************************/

#include "ommergelists_defs.h"
/*********************************************************
*
* The main program
*
********************************************************/
void ommergelists()
{

  //  double medc = 0.0;
  //  vector<double> z;
  //  vector<long> in;
  //  mlmcc(medc, z, in);



    string srcFile1 = stringParameter( "srclist1" ); // The first source-list file
    checkFileExists( srcFile1 );                     // Check the file exits

    int type1 = getType( srcFile1 );                 // SWS or observation?


    if(type1 != 0)
    {

        Sas::fatal("WrongType", "Can only process observation source-list files!");
        exit(0);

    }

    string srcFile2 = stringParameter( "srclist2" );  // The second source-list file
    if( srcFile1 == srcFile2)                         // Same files?
    {
        Sas::fatal("Identical source-list files- stopping", srcFile1);
       
    }


    checkFileExists( srcFile2 );

    int type2 = getType( srcFile2 );                 


    /********************************************************************
    *
    *
    * Both files have to be observation source-list files
    *
    **********************************************************************/

    if(type1 != 0)
    {
        Sas::fatal("WrongType", "Can only process observation source-list files!");
        exit(0);

    }

   
    /********************************************************************
    *
    * Want to check which file has a RAW_RATE column
    * File 2 should have this column, but if file 1 has it then 
    * need to swap file names
    *
    *********************************************************************/
    checkFiles(srcFile1, srcFile2);

    /*****************************************************************
    *
    *
    *
    ******************************************************************/
    string outSet = stringParameter( "outset"); // The output file for the merged source-list

     
    //    if (QFile::exists( outSet ) ) // Check to see if output file already exidts
     
    if ( dataSetServer->exists( outSet ) )
    {    
        Sas::warning("Output file exists", outSet );

    }
   
  
    /****************************************************************
    *
    * Load the source-parameters from the first  source-list
    * struct OmSourceParameters stores x, y, ra, dec, etc for a source
    *
    *****************************************************************/
    vector < OmSourceParameter > srcList1, srcList2;



    loadData(srcFile1, srcList1, type1);
    /****************************************************************
    *
    * Load the second source-list
    *
    *****************************************************************/
    loadData(srcFile2, srcList2, type2);

        
    /************************************************************************************
    *
    * Match the two lists using RA data
    * matchedindex1 indexes the matched sources in srclist1
    * matchedindex2 indexes the matched sources in srclist2
    * unmatchedindex1, 2 // Stores the indices of the unmatched indices in the two lists
    *
    ************************************************************************************/
    vector < OmSourceParameter > srcList3; 
    vector<int> matchedIndex1, matchedIndex2; 
    vector<int> unmatchedIndex1, unmatchedIndex2; 



    /***********************************************************************************
    *
    * NB on output the Right Ascension and declination coordinates of srclist2 
    * will have been corrected
    *
    *
    **********************************************************************************/
    matchLists( srcList1, srcList2, matchedIndex1, 
                matchedIndex2, unmatchedIndex1, unmatchedIndex2);

 
    /***********************************************************************************
    *
    * Draw some graphs comparing various properties of the matched sources from
    * the 2 lists
    *
    ***********************************************************************************/ 
    compareMatches( srcFile1, srcList1, matchedIndex1, srcFile2, srcList2, matchedIndex2 );

    /**************************************************************
    *
    * Create the output merged source-list file
    *
    **************************************************************/
    createMergedFile( matchedIndex1, unmatchedIndex1, matchedIndex2, unmatchedIndex2, srcList2 );



}
/**************************************************************
*
*
* Checks if the given file exists
*
**************************************************************/
void checkFileExists( string file )
{

  //    if (!QFile::exists( file ) )
    if ( !dataSetServer->exists( file ) )
    {
        
        Sas::fatal("File does not exist!", file);
      
    }

};

/*************************************************************
*
* Matches the sources in srclist1 and srclist2 using
* celestail coordinates
* matchedindex1 will contain an index list of the matched
* sources from srclist1
* Similarly matchedindex2 for srclist2
* unmatchedIndex1 will contain an index list of the unmatched
* source from srclist1
* Similarly unmatchedindex2 for srclist2
*
*
************************************************************/
void matchLists( vector < OmSourceParameter > &srcList1,
                 vector < OmSourceParameter > &srcList2,
                 vector < int > &matchedIndex1, 
                 vector < int > &matchedIndex2, 
                 vector < int > &unmatchedIndex1, 
                 vector < int > &unmatchedIndex2) 
               
{

    int size1 = (int)srcList1.size();

    errstr << "Matching source lists" 
           << " source list 1 size=" << srcList1.size()
           << " source list 2 size=" << srcList2.size()
	   << Sas::message(Msg::App, Msg::Sparse);


    /******************************************************************
    * Data for the source-lists is stored in a data structure
    *
    *******************************************************************/    
    vector<data> data1;

    for(int i = 0; i < size1; i++)
    {     

        data1.push_back(data(i, srcList1[i].rightascension * degreesToRadians, 
                                srcList1[i].declination * degreesToRadians, 
                                (float)1, (float)0, (int8)0) );
    }
    //    srcList2.resize(srcList1.size() );
       
    int size2 = (int)srcList2.size();

    vector<data> data2;

        
    for(int i = 0; i < size2; i++)
    {

      //      double offset = 2.5 / 3600.0;  // FOR TESTINGH
      //  srcList2[i].rightascension = srcList1[i].rightascension;  // FOR TESTING

      //	offset = 0.0;
      //  srcList2[i].declination = srcList1[i].declination + offset;  // FOR TESTING

        data2.push_back( data(i, srcList2[i].rightascension * degreesToRadians, 
                                 srcList2[i].declination * degreesToRadians, 
                                 (float)1, (float)0, (int8)0) );

    }


    /*********************************************************************
    *
    * Match the lists
    *
    *********************************************************************/ 
    double tolerance = realParameter( "tolerance" );   // The matching tolerance (arcsecs) is an input parameter
    MatchLists *match = new MatchLists(data1, data2);


    vector<data> matchedListData;
    vector<data> matchedCatData;
    int count = 0;


    double xOffset = 1e6, yOffset = 1e6; 
    double totalXOffset = 0.0, totalYOffset = 0.0;
    double tol = 0.05;

    int minSize = max(1, (int)data1.size() / 4);
    for(int i = 1; i <= 5; i++) // 5 Iterations
    {
       
 
        match->matchLists( tolerance);       // Do the matching

        /**********************************************************************
        *
        * Get the match for each source
        * If getallmatches - then gets all matches
        * if getmatches then just the nearest match
        *
        *********************************************************************/
        count = match->getMatches(tolerance, matchedListData, matchedCatData);
        if(count < minSize && tolerance < 5.0 ) 
	{
	    tolerance += 1.0;

	    continue;

        } 

        /*******************************************************************
	*
        *
	* Compute the RA and Dec mean offses between the matched data1
        * and matched data2 coordinates
        *
        *
        *******************************************************************/
        double dx = match->getRAMeanOffset();        
        double dy = match->getDecMeanOffset();

	//	cout << i << ") dx=" << dx << " dy=" << dy << endl;      
        if(abs(dx - xOffset) < tol && abs(dy - yOffset) < tol) break;

        xOffset = dx;
        yOffset = dy;
        totalXOffset += dx;
        totalYOffset += dy;

        /*************************************************************
	*
        * Remove the offsets 
        *
        **************************************************************/
        match->removeOffsets();

    } 

    delete match;

    if(count == 0) // No matches ?
    {
   
        Sas::warning("NoMatches", "No sources in the two lists have been matched" );

    }
    

    
    /****************************************************
    *
    * Correct srclist2 coordinates
    *
    *****************************************************/
    for(unsigned i = 0; i < srcList2.size(); i++)
    {
        srcList2[i].rightascension += (totalXOffset / 3600.0);
        srcList2[i].declination += (totalYOffset / 3600.0);
    }

    /**************************************************************
    *
    * Record which sources have been matched in both srclists
    * (RED regions)
    *
    **************************************************************/
    vector<bool> matched1(srcList1.size(), false );
    vector<bool> matched2(srcList2.size(), false );

    vector<string> colour;

    vector < OmSourceParameter > srcList3; // For the region file


    
    for(unsigned i = 0; i < matchedListData.size(); i++)
    {
        int k1 = matchedListData[i].index;
        matchedIndex1.push_back( k1 );
        matched1[ k1 ] = true;  // 
    
        int k2 = matchedCatData[i].index;
    
        matchedIndex2.push_back (k2);  // 
  
        matched2[ k2 ] = true;
       
	srcList3.push_back( srcList1[ k1 ] );
	colour.push_back( "red" );   // Matched sources

	//	cout << "Matched source in srclist1" << k1 << " with source " << k2 << " in srclist2\n"; 
    }

 
    /**********************************************************
    * The unmatched objects in srclist1
    * (GREEN regions) 
    ***********************************************************/
    for(unsigned i = 0; i < srcList1.size(); i++)
    {
        if ( !matched1[i] ) 
        {
	    unmatchedIndex1.push_back( i );
	    srcList3.push_back( srcList1[ i ] );
	    colour.push_back( "green" );
	    //   cout << "Index " << srcList1[i].number << endl;

        }
    }
   


    /**********************************************************
    * The unmatched objects in srclist2
    * (BLUE regions) 
    ***********************************************************/
    for(unsigned i = 0; i < srcList2.size(); i++)
    {
        if ( !matched2[i] )
        { 
	    unmatchedIndex2.push_back( i );
   
	    srcList3.push_back( srcList2[ i ] );
      
	    colour.push_back( "blue" );
	   

        }

    }

    /************************************************************
    * Create a region file to display the matches
    *
    **************************************************************/


    string regionFile = stringParameter( "regionfile");
    if(regionFile.length() > 2)
    {
        createRegionfile( regionFile, srcList3, colour, false);
    }


  
    
}
/***************************************************************
*
* Creates a region file showing the common sources to both
* sources-lists (red regions), those unique to the
* first one (green regions) and those unique to the second 
* one (green ones).
*
*
*
****************************************************************/
void createRegionfile(string file, const vector<OmSourceParameter> & sources, 
                                   const vector<string> &colour,   
                                   const bool showNumbers)
{

  //    cout << "creating region file\n";
    ofstream os( file.c_str() );

    os << "global select=0 edit=0 move=0 delete=1" << endl;

    float radius = 0.001;

    for (unsigned i = 0; i < sources.size(); i++) 
    {

        OmSourceParameter s = sources[i];    
     
        os << "fk5;circle("
	           << setprecision(12)  
		   << s.rightascension << ','
                   << setprecision(12)  
	
	           << s.declination    << ','    
	           << radius << ')'; 


 
 
	if ( showNumbers )
	{  
	  os  << " # text={" << s.number << "} color=" << colour[i];

	}
        else
	{

	    os  << " #  color=" << colour[i];


        }
	os << endl;


    }
   

    os.close();


}
/********************************************************************************
*
*
*
* Create the merged table using data from both of the input tables
*
* index is an index list of the matched sources in table 1
* index1 is an index list of the unmatched sources in table 1
* index2 is an index list of the unmatched sources in table 1
*
*   
*
* srcList2 will contain the corrected RA/DECS of source-list 2
*
********************************************************************************/

void createMergedFile( vector<int> &matchedIndex1, 
                       vector<int> &unmatchedIndex1, 
                       vector<int> &matchedIndex2, 
                       vector<int> &unmatchedIndex2,
                       vector < OmSourceParameter > &srcList2)
            
{
    
    errstr << "Merging tables" 
	   << Sas::message(Msg::App, Msg::Sparse);


    string srcFile1 = stringParameter( "srclist1" );
    string srcFile2 = stringParameter( "srclist2" );   
    string outSet = stringParameter( "outset");

    dataSetServer->copy(srcFile1, outSet);

    DataSet *dataSet = dataSetServer->open(outSet, Dal::Modify);
    if (!dataSet) return ;    

    Table *table = dataSet->table("SRCLIST");
  
    int size = table->rows();
  
    table->deleteRows(0, size);  // Delete all the rows


 


    int newSize = (int)( matchedIndex1.size() + unmatchedIndex1.size() + unmatchedIndex2.size());
    table->insertRows(0, newSize);






    DataSet *dataSet1 = dataSetServer->open(srcFile2, Dal::Read);

    Table *updateTable = dataSet1->table("SRCLIST");

    
    /*********************************************************
    * Add all the columns in the SRCLIST table of the second file
    * that are not present in the first one to the output file
    * newColumnList will store a list of the new columns
    *********************************************************/
    vector<string> newColumnList;
    addNewColumns(table, updateTable, newColumnList);



    /**************************************************************************
    *
    * Want to record which sourcelist set each value of each row in each column 
    * 0 Means that nothing set
    * 1 Means set by scrlist1
    * 2 means set by srclist2
    **************************************************************************/


    map<string, vector<int> > rowFilled;
    for(unsigned i = 0; i < table->columns(); i++)
    {
        string name = table->column(i)->name();
        rowFilled[name].resize(table->rows() );
        for(unsigned j = 0; j < table->rows(); j++)
	{
	    rowFilled[name][j] = 0; // Initialise everything to 0
        }

    }

  

    /*********************************************************
    *
    * Set every table value to NULL (except string columns)
    *
    *********************************************************/
    setNulls(table);

  

    dataSetServer->close(dataSet1); // Finished creating the output file


 



    dataSet1 = dataSetServer->open(srcFile1, Dal::Read);

    updateTable = dataSet1->table("SRCLIST");


    size = (int)matchedIndex1.size();


    vector<int> rowNums(size);
    for(int i = 0; i < size; i++)
        rowNums[i] = i;    

  

    /*************************************************************
    *
    *  Add the matched data from srclist1
    *  index is an index list of the matched sources in table 1
    * 
    *  The one indicates that srclist1 set a value in a row of a column
    *
    **************************************************************/
    fillTable( table, rowNums, updateTable, matchedIndex1, rowFilled, 1 );



  





    /****************************************************************
    *
    * Check to see if any of the matched sources in srclist2 have a 
    * raw-rate no greater than the maxrawrate value but a higher 
    * significance than the matched source in srclist1
    *
    *****************************************************************/

    int size1 = (int)unmatchedIndex1.size();
   
    /**********************************************
    *
    * Add the unmatched data from srclist1
    * index1 is an index list of the unmatched sources in table 1
    *
    *
    ***********************************************/
    rowNums.resize( size1 );
    for(int i = 0; i < size1; i++)
        rowNums[i] = size + i;


    /************************************************************************
    *
    * The one specifies that srclist 1 filled a particular value in a row
    * of a columns
    *
    ************************************************************************/  
    fillTable( table, rowNums, updateTable, unmatchedIndex1, rowFilled, 1 );



    /*******************************************************************
    *
    * Finished adding data from SRCLIST1 
    *
    ********************************************************************/
    dataSetServer->close(dataSet1); // Finished with the file

 

    /*********************************************************************
    *
    * Load the table from srcFile2
    *
    **********************************************************************/
    dataSet1 = dataSetServer->open(srcFile2, Dal::Read);
    updateTable = dataSet1->table("SRCLIST");

 
    /****************************************************************
    *
    * Check to see if any of the matched sources in srclist2 have a 
    * raw-rate no greater than the maxrawrate value but a higher 
    * significance than the matched source in srclist1
    *
    *****************************************************************/
    checkSignificances(table, updateTable, matchedIndex1, matchedIndex2, rowFilled);





    /*********************************************************
    *
    * Add the matched sources from srclist2
    * index2 is an index list of the unmatched sources in table 2
    *
    **********************************************************/ 
    int size2 = (int)matchedIndex2.size();

    rowNums.resize( size2 );
    for(int i = 0; i < size2; i++)
        rowNums[i] = i;    


    /****************************************************************
    *
    * matchedIndex2 is an index list of the matched sources in table 2
    * Add the matched sources from srclist2 to the table
    *
    *****************************************************************/
    fillTable( table, rowNums, updateTable, matchedIndex2, rowFilled, 2 );







    /*********************************************************
    *
    * Add the unmatched sources from srclist2
    * index2 is an index list of the unmatched sources in table 2
    *
    **********************************************************/
    size2 = (int)unmatchedIndex2.size();


    rowNums.resize( size2 );
    for(int i = 0; i < size2; i++)
        rowNums[i] = size + size1 + i;    


    /****************************************************************
    * index2 is an index list of the unmatched sources in table 2
    *
    *****************************************************************/
    fillTable( table, rowNums, updateTable, unmatchedIndex2, rowFilled, 2 );



 
    dataSetServer->close(dataSet1); // Finished with the update file





    /*************************************************************
    *
    * Need to add a column for each filter to say which 
    * RATE came from a MOSAICED sky image
    *
    **************************************************************/
    addSkyImageColumns(table, rowFilled);    




    /*****************************************************
    * Need to copy RA, DEC columns to RA_CORR/DEC_CORR columns
    * in combined list for the new sky-image sources
    *
    *****************************************************/
    setRADecColumns( table, rowNums);




    /**********************************************************************
    *
    * Need to set the RA_CORR and DEC_CORR columns in the output table
    * for those unmatched sources in the update table
    * The coordinates will have been modified in the matching process
    * ie- aligned with those in the first source-list file
    ***********************************************************************/
    setCorrectedCoordinates(table, srcList2, rowNums, unmatchedIndex2);




    /*************************************************************************
    *
    *
    *
    *************************************************************************/
    checkNewRateColumns(table, rowNums);





    /****************************************************
    * Remove RAW_RATE column
    *
    *****************************************************/
    removeColumns(table);


    /****************************************************
    *
    * Sort the table into ascending RA order
    *
    ****************************************************/
    sortTable( &table );





    dataSetServer->close(dataSet); // Finished with the output file
 
 
}





/************************************************************
*
*
*  Updates the columns in one table with data from the columns
*  in another table with the same column names
*
*
* 
*************************************************************/
void fillTable( Table * table,           // The merged output table  
                vector<int> &rowNums,    // The list of row numbers to be updates
                Table *updateTable,      // The update table
                vector<int> &updateNums, // The list of row numbers in the update table 
                map<string, vector<int> > &filled, // Records what srclist filled a particular row in a column 
                int fillNumber )                   // The number to set filled to if its value is zero to begi with 
{

    /*************************************************************
    *
    * Loop through all the columns of table 1
    *
    *************************************************************/
    for( unsigned i = 0; i < table -> columns(); ++i )
    {
        Column * col = table -> column( i );
        string columnName = col->name();
         
	/**********************************************************
	* See if this column exists in the update table
        *
        **********************************************************/
        if ( !updateTable -> hasColumn( columnName) )
        {
	    continue; // No such column in the update table- just skip  

        }

	int kStart = updateTable->columnNumber( columnName ); 
     
        /****************************************************************
	*
	* Update the column col with the update column data
        *
        ****************************************************************/        
  	Column * updateCol = updateTable -> column( kStart );
     

     	ColumnData * c = col -> data();
     	ColumnData * c1 = updateCol -> data();
     			
	switch( col -> dataType() )  // Need to know what the column data type is
	{
	    case Column::Bool :
     	          
		break;
	    case Column::Int8 :
		fillData( c -> int8Data(), rowNums, c1 -> int8Data(), updateNums, filled[columnName], fillNumber);
		break;

	    case Column::Int16 :
		fillData( c -> int16Data(), rowNums, c1 -> int16Data(), updateNums, filled[columnName], fillNumber);
	
		break;
	    case Column::Int32 :
		
	        fillData( c -> int32Data(), rowNums, c1 -> int32Data(), updateNums, filled[columnName], fillNumber);
	
		break;
	    case Column::Real32 :
		fillData( c -> real32Data(), rowNums, c1 -> real32Data(), updateNums, filled[columnName], fillNumber );
		break;

	    case Column::Real64 :
		fillData( c -> real64Data(), rowNums, c1 -> real64Data(), updateNums, filled[columnName], fillNumber );
		
		break;
	    case Column::DString :   
	    {
	        int width = c->parent()->elements(); 	  
	        fillData( c -> stringData(), rowNums, c1 -> stringData(), updateNums, width, filled[columnName], fillNumber );
	
		break;
            }
	    default :
		break;
	} // switch

	delete c;
        delete c1;



    } // for

}

   
/********************************************************************************
*
* Basically for testing purposes
*
*
*
*********************************************************************************/
void compareMatches( vector<data> &matchedListData, vector<data> &matchedCatData)
{


    if(matchedListData.size() == 0) return;

    double sum = 0.0;
    vector<double> distance, dx, dy;

    for(unsigned i = 0; i < matchedListData.size(); i++)
    {

        double d = matchedListData[i].arcLength( matchedCatData[i] );
        sum += d; 
        distance.push_back( d );
        double delta = (matchedListData[i].ra - matchedCatData[i].ra) * cos(matchedListData[i].dec * degreesToRadians);
        dx.push_back(delta * 3600.0);
        dy.push_back( (matchedListData[i].dec - matchedCatData[i].dec) * 3600.0);

    }

    cout << "Mean separation between matched sources=" << sum / double(matchedListData.size()) << " arcsecs\n";
    cout << "Minimum separation=" << MinVal(distance) << " maximum=" << MaxVal(distance) 
         << " median=" << Median(distance) << endl;

    cout << "0.025 percentile=" << Percentile( distance, 0.025) << " 0.975 percentile=" <<  Percentile( distance, 0.975) << endl;


    cout << "Mean RA offset between matched sources=" << Mean(dx) << " arcsecs\n";
    cout << "Minimum offset=" << MinVal(dx) << " maximum=" << MaxVal(dx) 
         << " median=" << Median(dx) << endl;

    cout << "0.025 percentile=" << Percentile( dx, 0.025) << " 0.975 percentile=" <<  Percentile( dx, 0.975) << endl;



    cout << "Mean Dec offset between matched sources=" << Mean(dy) << " arcsecs\n";
    cout << "Minimum offset=" << MinVal(dy) << " maximum=" << MaxVal(dy) 
         << " median=" << Median(dy) << endl;

    cout << "0.025 percentile=" << Percentile( dy, 0.025) << " 0.975 percentile=" <<  Percentile( dy, 0.975) << endl;

}
/**********************************************************************
*
*
* This is for comparing properties between the two lists of matched sources
*
***********************************************************************/
void compareMatches( vector < OmSourceParameter > &srcList1,
                     vector < int >               &matchedIndex1,
                     vector < OmSourceParameter > &srcList2,
                     vector < int >               &matchedIndex2)

 {
    if(matchedIndex1.size() < 3) return;



    /**************************************************************
    *
    * Initialise pgplot
    *
    ***************************************************************/
    string plotFile = stringParameter("plotfile");
    if(plotFile.length() < 2) return;
        

        cout << "Line 1017\n";
    plotFile = plotFile + "/VC"; 
    int ok =  cbpgopen((char *)plotFile.c_str());

    if (ok == 0) {

        cout << "Could not open pgplot device\n";
        exit(0);

    } 

  cout << "Line 1028\n";











    //    cout << "compare matches line 851\n";

    cout << "\nComparing matched corrected count rates and significances\n\n";

    cout << setw(6) << "Num" << setw(12) << "E_FLAG1" << setw(12) << "E_FLAG2" << setw(12) << "Raw1" << setw(12) << "Raw2" << setw(15) << "Raw1-Raw2" 
	 << setw(12) << "Sig1" << setw(12) << "Sig2" << setw(15) << "Sig1-Sig2" << setw(12) << "Corr1" << setw(12) << "Corr2" << endl; 


    int size = (int)matchedIndex1.size();
    vector<float> diff1( size );
    vector<float> diff2( size );
  
    vector<float> rate1(size), rate1Err(size);
    vector<float> rate2(size), rate2Err(size);
    vector<float> corrRate1(size), corrRate1Err(size);
    vector<float>corrRate2(size), corrRate2Err(size);

  
    vector<float> logRate1(size), logRate1Err(size);
    vector<float> logRate2(size), logRate2Err(size);
    vector<float> logCorrRate1(size), logCorrRate1Err(size);
    vector<float> logCorrRate2(size), logCorrRate2Err(size);


    vector<float> rateRatio(size);
    vector<float> corrRateRatio(size);


    vector<float> sig1(size), sig2(size);
    vector<double> ra1(size), dec1(size), poserr1(size);
    vector<double> ra2(size), dec2(size), poserr2(size);

    vector<float> semia1(size), semia2(size);
  


    vector<int> colourIndex( size, 2);



    for(int i = 0; i < size; i++)
    {


        int k1 = matchedIndex1[i];
        int k2 = matchedIndex2[i];
	if( srcList1[k1].extensionflag == 0 &&  srcList2[k2].extensionflag == 0 ) 
	{
	    colourIndex[i] = 4;
 
        } 
        else
	{
	    colourIndex[i] = 2;
 
        }

 
        semia1[i] = srcList1[k1].semia;
        semia2[i] = srcList2[k2].semia;
   

        ra1[i] = srcList1[k1].rightascension;
        ra2[i] = srcList2[k2].rightascension;
        poserr1[i] = srcList1[k1].positionerror / 3600.0;
        poserr2[i] = srcList2[k2].positionerror / 3600.0;

        dec1[i] = srcList1[k1].declination;
        dec2[i] = srcList2[k2].declination;



        diff1[i] = srcList1[k1].rawcountrate - srcList2[k2].rawcountrate;
        diff2[i] = srcList1[k1].significance - srcList2[k2].significance;

        rate1[i] = srcList1[k1].rawcountrate;
        logRate1[i] = log10( srcList1[k1].rawcountrate);

        corrRate1[i] = srcList1[k1].corrcountrate;
        logCorrRate1[i] = log10( srcList1[k1].corrcountrate);


        sig1[i] = log10( srcList1[k1].significance );
        sig2[i] = log10( srcList2[k2].significance );

        rate1Err[i] = srcList1[k1].rawcountrateerror;
	rate1Err[i] = 0.0;
	corrRate1Err[i] = 0.0;
   
        rate2[i] = srcList2[k2].rawcountrate;
        logRate2[i] = log10(srcList2[k2].rawcountrate);
        rateRatio[i] = rate1[i] / rate2[i];


        rate2Err[i] = srcList2[k2].rawcountrateerror;
        corrRate2[i] = srcList2[k2].corrcountrate;
        logCorrRate2[i] = log10( srcList2[k2].corrcountrate);



	rate2Err[i] = 0.0;
   	corrRate2Err[i] = 0.0;
   
        corrRateRatio[i] = corrRate1[i] / corrRate2[i];

      
        cout << setw(5) << i << setw(12) << (int)srcList1[k1].extensionflag << setw(12) << (int)srcList2[k2].extensionflag 
             <<   setw(12) << srcList1[k1].rawcountrate << setw(12) << srcList2[k2].rawcountrate << setw(15) << diff1[i] 
	     << setw(12) << srcList1[k1].significance << setw(12) << srcList2[k2].significance << setw(15) << diff2[i]   
             << setw(12) << srcList1[k1].corrcountrate << setw(12) << srcList2[k2].corrcountrate << endl; 



    }


    cout << "\nRaw11-Raw2 statistics\n";
    cout << "Mean=" << Mean(diff1) << " median=" << Median(diff1) << endl;
    cout << "0.025 percentile=" << Percentile( diff1, 0.025) << " 0.975 percentile=" <<  Percentile( diff1, 0.975) << endl;

    cout << "\nSig1-Sig2 statistics\n";
    cout << "Mean=" << Mean(diff2) << " median=" << Median(diff2) << endl;
    cout << "0.025 percentile=" << Percentile( diff2, 0.025) << " 0.975 percentile=" <<  Percentile( diff2, 0.975) << endl;




    /************************************************************************
    *
    * Right Ascension plot
    *
    *
    ************************************************************************/
    
    cbpgsubp(2, 2);
    cbpgpanl(1, 1);

    float xMin = floor( MinVal( ra1 ) * 10.0) / 10.0;
    float xMax = ceil( MaxVal( ra1 ) * 10.0) / 10.0;;
 
    float yMin = floor( MinVal( ra2 ) * 10.0) / 10.0;
    float yMax = ceil( MaxVal( ra2 ) * 10.0) / 10.0;;
 
    xMin = min( xMin, yMin);
    xMax = max( xMax, yMax);
    yMin = xMin;
    yMax = xMax;


    drawGraph( ra1, poserr1, 
               ra2, poserr2,
               colourIndex,
               (char *)"", 
               (char *)"Unrotated Right Ascension (degrees)", 
               (char *)"Mosaiced Right Ascension (degrees)", 
           
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, 
               true);

    cbpgsci(2);
    cbpgmove(xMin, yMin);
    cbpgdraw(xMax, yMax);

    cbpgsci(1);
 
    cbpgpanl(2, 1);



    vector<double> raDiff = difference(ra1, ra2, 3600.0);
    

    xMin = floor( MinVal( raDiff ) * 10.0 ) / 10.0;
    xMax = ceil( MaxVal( raDiff ) * 10.0) / 10.0;
 
    xMin = min(xMin, -xMax);
    xMax = abs(xMin);

    double binWidth = max(0.1, (xMax - xMin) / 20.0);

    Histogram( raDiff, binWidth, 
	       (char *)"\\gD RA (arcsecs)", (char *)"", 
	       0.15, 0.9, 0.15, 0.9, xMin, xMax);


    cbpgsls(2);
    cbpgsci(2);



    cbpgmove(0.0, 0.0);
    cbpgdraw(0.0, 100000.0);
    cbpgsls(1);
    cbpgsci(1);
    /************************************************************************
    *
    * Declination plot
    *
    ************************************************************************/
    cbpgpanl(1, 2);


    xMin = floor( MinVal( dec1 ) * 10.0) / 10.0;
    xMax = ceil( MaxVal( dec1 ) * 10.0) / 10.0;;
 
    yMin = floor( MinVal( dec2 ) * 10.0) / 10.0;
    yMax = ceil( MaxVal( dec2 ) * 10.0) / 10.0;;
 
    xMin = min( xMin, yMin);
    xMax = max( xMax, yMax);
    yMin = xMin;
    yMax = xMax;


    drawGraph( dec1, poserr1, 
               dec2, poserr2,
               colourIndex,
               (char *)"", 
               (char *)"Unrotated Declination (degrees)", 
               (char *)"Mosaiced Declination (degrees)", 
           
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, 
               true);



  


    cbpgsci(1);
    cbpgmove(xMin, yMin);
    cbpgdraw(xMax, yMax);


    /*******************************************************************
    *
    * Next graph
    *
    *******************************************************************/



    cbpgpanl(2, 2);


    vector<double> decDiff = difference(dec1, dec2, 3600.0);

    xMin = floor( MinVal( decDiff ) );
    xMax = ceil( MaxVal( decDiff ) );
 
    xMin = min(xMin, -xMax);
    xMax = abs(xMin);

    binWidth = max(0.1, double(xMax - xMin) / 20.0);

    
    Histogram( raDiff, binWidth, 
	       (char *)"\\gD Dec (arcsecs)", (char *)"", 
	       0.15, 0.9, 0.15, 0.9, xMin, xMax);



    cbpgsls(2);
    cbpgsci(2);


    cbpgmove(0.0, 0.0);
    cbpgdraw(0.0, 100000.0);

    cbpgsls(1);
    cbpgsci(1);


    cbpgpage();


    /********************************************************************
    * Next page
    *
    *********************************************************************/



    /************************************************************
    * Raw rates
    *
    ************************************************************/


    cbpgpanl(1, 1);






    xMin = floor( MinVal( logRate1 ) * 10.0) / 10.0;
    xMax = ceil( MaxVal( logRate1 ) * 10.0) / 10.0;;
 
    yMin = floor( MinVal( logRate2 ) * 10.0) / 10.0;
    yMax = ceil( MaxVal( logRate2 ) * 10.0) / 10.0;;
 
    xMin = min( xMin, yMin);
    xMax = max( xMax, yMax);
    yMin = xMin;
    yMax = xMax;




    drawGraph( logRate1, rate1Err, 
               logRate2, rate2Err,
               colourIndex,
               (char *)"", 
               (char *)"Log\\d10\\d(Unrotated raw count-rate) counts/sec", 
               (char *)"Log\\d10\\d(Mosaiced raw count-rate) counts/sec", 
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, 
               true);
    cbpgsci(1);
    cbpgmove(xMin, yMin);
    cbpgdraw(xMax, yMax);


    vector<float> ratio;
    for(unsigned i = 0; i < corrRate1.size(); i++)
    {
      ratio.push_back( corrRate1[i] / corrRate2[i] );

    }

    plotStatistics(ratio, "Rate1/Rate2");



    /*******************************************************************
    *
    * Next graph
    *
    *******************************************************************/
    cbpgpanl(2, 1);



    xMin = floor( MinVal( logRate1 ) * 10.0) / 10.0;
    xMax = ceil( MaxVal( logRate1 ) * 10.0) / 10.0;;
 
    yMin = floor( MinVal( rateRatio ) * 10.0) / 10.0;
    yMax = ceil( MaxVal( rateRatio ) * 10.0) / 10.0;;
 
 
  



        drawGraph( logRate1, rate1Err, 
                 rateRatio, rate2Err,
               colourIndex,
               (char *)"MOSAICED IMAGING TESTS", 
               (char *)"Log(Unrotated image raw count-rate) counts/sec", 
               (char *)"Unrotated / mosaiced raw count-rate", 
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, 
               false);
    cbpgsci(1);
    cbpgmove(xMin, 1.0);
    cbpgdraw(xMax, 1.0);

   
    /*****************************************************************
    *
    *
    ******************************************************************/
    cbpgpanl(2, 1);

  
    xMin = floor( MinVal( logCorrRate1 ) * 10.0) / 10.0;
    xMax = ceil( MaxVal( logCorrRate1 ) * 10.0) / 10.0;;
 
    yMin = floor( MinVal( logCorrRate2 ) * 10.0) / 10.0;
    yMax = ceil( MaxVal( logCorrRate2 ) * 10.0) / 10.0;;
 
    xMin = min( xMin, yMin);
    xMax = max( xMax, yMax);
    yMin = xMin;
    yMax = xMax;


    drawGraph( logCorrRate1, corrRate1Err, 
               logCorrRate2, corrRate2Err,
               colourIndex,
               (char *)"", 
               (char *)"Log\\d10\\d(Unrotated corrected count-rate) counts/sec", 
               (char *)"Log\\d10\\d(Mosaiced corrected count-rate) counts/sec", 
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, 
               true);
    cbpgsci(1);
    cbpgmove(xMin, yMin);
    cbpgdraw(xMax, yMax);



    /********************************************************************
    *
    * Next page
    *
    ********************************************************************/


    //    cpgpage();

    cbpgpanl(1, 2);



    xMin = floor( MinVal( sig1 ) * 10.0) / 10.0;
    xMax = ceil( MaxVal( sig1 ) * 10.0) / 10.0;;
 
    yMin = floor( MinVal( sig2 ) * 10.0) / 10.0;
    yMax = ceil( MaxVal( sig2 ) * 10.0) / 10.0;;
 
    xMin = min( xMin, yMin);
    xMax = max( xMax, yMax);
    yMin = xMin;
    yMax = xMax;



   
    drawGraph( sig1, rate1Err, 
               sig2, rate2Err,
               colourIndex,
               (char *)"", 
               (char *)"Log\\d10\\d(Unrotated significance)", 
               (char *)"Log\\d10\\d(Mosiaced significance)", 
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, 
               true);
    cbpgsci(1);
    cbpgmove(xMin, yMin);
    cbpgdraw(xMax, yMax);
    cbpgpage();



    cbpgpanl(2, 2);



    xMin = floor( MinVal( semia1 ) * 10.0) / 10.0;
    xMax = ceil( MaxVal( semia1 ) * 10.0) / 10.0;;
 
    yMin = floor( MinVal( semia2 ) * 10.0) / 10.0;
    yMax = ceil( MaxVal( semia2 ) * 10.0) / 10.0;;
 
    xMin = min( xMin, yMin);
    xMax = max( xMax, yMax);
    yMin = xMin;
    yMax = xMax;

    vector<float> error(semia1.size(), 0.0);


    drawGraph( semia1, error, semia2, error,
               colourIndex,
               (char *)"", 
               (char *)"Unrotated semi-major axis (arcsec)", 
               (char *)"Mosiaced semi-major axis  (arcsec)", 
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, 
               true);
    cbpgsci(1);
    cbpgmove(xMin, yMin);
    cbpgdraw(xMax, yMax);


    cbpgclos();


}
/*********************************************************************
*
* Return 0 if the file is a observation source-list file- 1 otherwise
*
********************************************************************/
int getType( string file )
{


    DataSet *dataSet = dataSetServer->open( file, Dal::Read);

    if (!dataSet)
    {


    }    
 
    
    string filter = dataSet->attribute( "FILTER" )->asString();

    int type = 0;
    if ( filter == "n/a" || filter == "N/A") // Observation source list file
    {
        type = 0;
    }
    else
    {
        type = 1;

    }

    dataSetServer->close(dataSet); // Finished with the file

    return type;
}
/**********************************************************************
*
* Loads data from an OM observation source-list file
*
*
*
***********************************************************************/
void loadObsSourceParameters( string file,   vector < OmSourceParameter > &srcList)
{

    DataSet *dataSet = dataSetServer->open( file, Dal::Read);

    if (!dataSet)
    {


    }    
 
    const Table *table = dataSet->table( "SRCLIST", Dal::Read);

    /**********************************************************************
    *
    *
    *
    **********************************************************************/
    vector<double> data;
    int size = 0;

    if(table->hasColumn("RA_CORR"))
    {
        if ( !LoadVectorDoubleColumnData( table, "RA_CORR", size, data))
        {

            dataSetServer->close(dataSet); // Finished with the file

            Sas::fatal("Error loading column from table-", "RA_CORR");

        }


        srcList.resize(size);
        for(int i = 0; i < size; i++)
            srcList[i].rightascension = data[i];


        data.clear();
        if ( !LoadVectorDoubleColumnData( table, "DEC_CORR", size, data))
        {
            dataSetServer->close(dataSet); // Finished with the file
            Sas::fatal("Error loading column from table-", "DEC_CORR");


        }
        for(int i = 0; i < size; i++)
            srcList[i].declination = data[i];
    }
    else
    {

    
        if ( !LoadVectorDoubleColumnData( table, "RA", size, data))
        {

            dataSetServer->close(dataSet); // Finished with the file
            Sas::fatal("Error loading column from table-", "RA");


        }


        srcList.resize(size);
        for(int i = 0; i < size; i++)
            srcList[i].rightascension = data[i];

        data.clear();
        if ( !LoadVectorDoubleColumnData( table, "DEC", size, data))
        {

            Sas::fatal("Error loading column from table-", "DEC");


        }
        for(int i = 0; i < size; i++)
            srcList[i].declination = data[i];
   
    }


    /*******************************************************************
    *
    * Need to get some significances
    *
    ********************************************************************/
    map<string, vector<float> > significanceMap;


    for(int i = 0; i < 7; i++)
    {
        string filter = OmFilter::filter[i];
        vector<float> significance(srcList.size(), -1.0);
 

        string name = OmFilter::filter[i] + "_SIGNIF";
        if(table->hasColumn(name) )
	{
	  //    cout << "Loading significances from column " << name << endl;

	    //   vector<float> data, data1, data2;
            vector<float> data;

            int size = 0;
            if ( !LoadVectorFloatColumnData( table, name, size, data))
            {
            }
            significanceMap[filter] = significance;


            for(int i = 0; i < size; i++)
	    {
	        if (!isNull( data[i] ) )
                    significanceMap[filter][i] = data[i] ;
            }
   
                     
           
        }
        name = OmFilter::filter[i] + string("_SIGNIFICANCE");
        if(table->hasColumn(name) )
	{
	  //  cout << "Loading significances from column " << name << endl;

	    vector<float> data;
            int size = 0;
            if ( !LoadVectorFloatColumnData( table, name, size, data))
            {


            }
            significanceMap[filter] = significance;
      
            for(int i = 0; i < size; i++)
	    {     
                if (!isNull( data[i] ) )
                    significanceMap[filter][i] = data[i];
       
            } 
     
        }
    }



    dataSetServer->close(dataSet); // Finished with the file
 
    map<string, vector<float> >::iterator iter;


    vector<float> significance(srcList.size(), -1.0);
 
    for(unsigned i = 0; i < srcList.size(); i++)
      srcList[i].significance = -1.0;
    

    for(iter =  significanceMap.begin(); iter != significanceMap.end(); ++iter)
    {
        vector<float> data = iter->second;
        for(unsigned i = 0; i < data.size(); i++)
	    srcList[i].significance = max(   srcList[i].significance, data[i] );
          

    }

 


}

/*************************************************************
*
* Only want to include in the output source-list file
* those unique sources in the second input-source list
* that have a significance less than or equal to maxsignificance
*
*
************************************************************/
void compareMatchedSignificances( vector < OmSourceParameter > &srcList1,
                 vector < OmSourceParameter > &srcList2,
                 vector < int > &matchedIndex1, 
                 vector < int > &matchedIndex2, 
                 vector < int > &unmatchedIndex2) 
               
{


    if(matchedIndex1.size() == 0) return; 

    vector < int > matchedIndex11; 
    vector < int > matchedIndex21; 
  
               
    float maxSignificance = realParameter("maxsignificance");


    /**********************************************************
    *
    * Loop through all the matched sources and select the
    * matched source with the highest significance
    *
    ***********************************************************/


    for(unsigned i = 0; i < matchedIndex1.size(); i++) 
    {

        int index1 =  matchedIndex1[i];
        int index2 =  matchedIndex2[i];
       

        if( srcList1[index1].significance >= maxSignificance)
	{


	    matchedIndex11.push_back( index1 ); 
            matchedIndex21.push_back( index2 ); 
         
            continue;
        } 
      
        
        if( srcList2[index2].significance > srcList1[index1].significance )
	{

            unmatchedIndex2.push_back(  index2 ); 
            continue;

        }
        
	matchedIndex11.push_back( index1 ); 
        matchedIndex21.push_back( index2 ); 
        
  
    }

 
    matchedIndex1 = matchedIndex11; 
    matchedIndex2 = matchedIndex21; 
              
              
}
/**************************************************************
*
*
**************************************************************/
void removeUnusedRows(Table *table, vector<bool> &used)
{

    for(int i = (int)table->rows() - 1; i >= 0; i--)
    {
        if ( !used[i] )
        {  
            table->deleteRows(i, 1);  // Delete all the rows
  
        }
    }

}
/*******************************************************************
*
* Adds a new column to the table
* column contains information about the column to be added to the table
* table contains information about the table 
*******************************************************************/
void addColumn(Column *column, Table *table)
{


         	
    unsigned size =  table->rows();  // The number of rows in the table

    switch( column -> dataType() )  // Find out what the column data-type is
    {
        case Column::Bool :

	    // Add the column 
   
     	    table->addColumn( column->name(), Column::Bool, column->label(), column->units());
          
	    break;
        case Column::Int8 :  // Integer 8
	{	  
	    // Add te column to the table and set all the data values to null 
            Column *newColumn = table->addColumn( column->name(), Column::Int8, column->label(), column->units());

         
            ColumnData *c = newColumn->data();


            newColumn -> nullValue( 128 );


	    for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;
  
 
	    break;
        }
        case Column::Int16 :  // Integer 16
	{  
	    
            // Add the column to the table and set all the data values to null         
            Column *newColumn = table->addColumn( column->name(), Column::Int16, column->label(), column->units());

            ColumnData *c = newColumn->data();

   
            newColumn -> nullValue( 32767 );

            for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;
 
	    break;
        }
        case Column::Int32 :  // Integer 32
	{

           // Add the column to the table and set all the data values to null 
         
	    Column *newColumn = table->addColumn( column->name(), Column::Int32, column->label(), column->units());
   
            newColumn -> nullValue( 32767 );

            ColumnData *c = newColumn->data();

            for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;
 
	    break;
        }
        case Column::Real32 : // Float
	{
            // Add the column to the table and set all the data values to null 	    
            Column *newColumn = table->addColumn( column->name(), Column::Real32, column->label(), column->units() );


          
            ColumnData *c = newColumn->data();

            for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;  
	    break;
	}
        case Column::Real64 :  // Double
	{
            // Add the column to the table and set all the data values to null 
         

	    Column *newColumn = table->addColumn( column->name(), Column::Real64, column->label(), column->units()  );
          
            ColumnData *c = newColumn->data();


            for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }

            delete c;
 
	    break;
          }
	case Column::DString :    // String
	{

            // Add the column to the table and set all the data values to null 
         

            ColumnData *c = column->data();

   
	    int width = c->parent()->elements();
            longVec size(2);
	    size[0] = width; 
            size[1] = table->rows();	  
	    table->addColumn( column->name(), Column::DString, column->label(), column->units(), size );
      
            delete c;
	    break;
        }

	default :
	    break;
	} // switch


    
}

/**************************************************************************
*
*
* Need to check to see if all the columns in update table exist in the 
* table- if not add them to the table
*
*
**************************************************************************/
void addNewColumns(Table *table, Table *updateTable, 
                   vector<string> &newColumnList)
{
 
    for( unsigned int i = 0; i < updateTable -> columns(); ++i )
    {
      
        string colName = updateTable -> column( i )->name();

        if( !table->hasColumn( colName ) )
	{

	    addColumn( updateTable->column(i), table );
	    newColumnList.push_back( colName );
    
        }
    
    }   
}	

/**********************************************************************
*
*
* This is for comparing properties between the two lists of matched sources
*
*
***********************************************************************/
void compareMatches( string file1,
                     vector < OmSourceParameter > &srcList1,
                     vector < int >  &matchedIndex1,
                     string file2,
                     vector < OmSourceParameter > &srcList2,                 
                     vector < int >   &matchedIndex2)

{




    if (matchedIndex1.size() < 3) return;


    errstr << "Drawing graphs to compare matched-source properties" 
	   << Sas::message(Msg::App, Msg::Sparse);

  
 

 



    /**************************************************************
    *
    * Initialise pgplot
    *
    ***************************************************************/
    


    //   cout << "Line 2033\n";
    string plotFile = stringParameter("plotfile");
    if(plotFile.length() < 2) return;
        
    plotFile = plotFile + "/VC"; 

    //    cout << "Line 2039 plotfile=" << plotFile << endl;

    int ok =  cbpgopen((char *)plotFile.c_str());

    // cout << "Line 2043\n";


    if (ok == 0) {

        cout << "Could not open pgplot device\n";
        exit(0);

    } 

  
  
    // cout << "Line 2053\n";

    /*******************************************************************
    *
    *
    *
    *******************************************************************/

    DataSet *dataSet1 = dataSetServer->open(file1, Dal::Read);

    string obsId = dataSet1->attribute("OBS_ID")->asString();


    Table *table1 = dataSet1->table("SRCLIST");

    int size = (int)table1->rows();    



    /****************************************************************************
    *
    * Load the RAs and Decs from the first table
    *
    *****************************************************************************/

    vector<double> ra1(size), dec1(size);
    for(int i = 0; i < size; i++)
    {
      ra1[i] = srcList1[i].rightascension;
      dec1[i] = srcList1[i].declination;


    }

    DataSet *dataSet2 = dataSetServer->open(file2, Dal::Read);
    Table *table2 = dataSet2->table("SRCLIST");
    size = (int)table2->rows();

    /****************************************************************************
    *
    * Load the RAs and Decs from the second table
    *
    *****************************************************************************/
    vector<double> ra2(size), dec2(size);
    
    for(int i = 0; i < size; i++)
    {
      ra2[i] = srcList2[i].rightascension;
      dec2[i] = srcList2[i].declination;


    }
 
    // cout << "Line 2106\n";


    vector<double> ra3, ra4, dec3, dec4;
    for(unsigned i = 0; i < matchedIndex1.size(); i++)
    {
        int k1 = matchedIndex1[i];
        int k2 = matchedIndex2[i];
        ra3.push_back(ra1[k1] );
        dec3.push_back(dec1[k1]);
        ra4.push_back(ra2[k2]);
        dec4.push_back(dec2[k2]);
 
    }

  


    cbpgsubp(1, 2);
    cbpgpanl(1, 1);


    //   cout << "Line 2128\n";

    string title = obsId + " Right Ascension comparison";
 

    vector<double> raDiff = difference(ra3, ra4, 3600.0);
    for(unsigned i = 0; i < raDiff.size(); i++)
        raDiff[i] *= cos(dec3[i] * degreesToRadians);

    double xMin1 = floor( MinVal( raDiff ) * 10.0) / 10.0;
    double xMax1 = ceil( MaxVal( raDiff ) * 10.0 ) / 10.0;
 
    xMin1 = min(xMin1, -xMax1);
    xMax1 = abs(xMin1);



   
    vector<double> decDiff = difference(dec3, dec4, 3600.0);



    double xMin2 = floor( MinVal( decDiff ) * 10.0) / 10.0;
    double xMax2 = ceil( MaxVal(  decDiff ) * 10.0) / 10.0;
 
    xMin2 = min(xMin2, -xMax2);
    xMax2 = abs(xMin2);


    double xMin = min(xMin1, xMin2);
    double xMax = max(xMax1, xMax2);



    //  cout << "Line 2162\n";



    double binWidth = (xMax - xMin) / 20.0;
    binWidth = max(0.1, ceil(binWidth * 10.0) / 10.0);

    binWidth = 0.1;

    if(xMax > 1.0)
    {
        binWidth = 0.1;
    }
    else
    {
        binWidth = 0.1;
    }


    //      cout << "Line 2181\n";


    Histogram( raDiff, binWidth, 
	       (char *)"\\gD RA x cos(\\gd) (arcsecs)", (char *)title.c_str(), 
	       0.15, 0.9, 0.15, 0.9, xMin, xMax);



    //     cout << "Line 2190\n";


    cbpgsls(2);
    cbpgsci(2);


    cbpgmove(0.0, 0.0);
    cbpgdraw(0.0, 100000.0);
    cbpgsls(1);
    cbpgsci(1);


    //  cout << "Line 2204\n";

    /************************************************************************
    *
    * Declination plot
    ************************************************************************/
    title = obsId + " Declination comparison";

    cbpgpanl(1, 2);

 
    binWidth = (xMax - xMin) / 20.0;
    binWidth = max(0.1, ceil(binWidth * 10.0) / 10.0);
    if(xMax > 1.0)
    {
        binWidth = 0.1;
    }
    else
    {
        binWidth = 0.1;
    }

    Histogram( raDiff, binWidth, 
	       (char *)"\\gD Dec (arcsecs)", (char *)title.c_str(), 
	       0.15, 0.9, 0.15, 0.9, xMin, xMax);

    cbpgsls(2);
    cbpgsci(2);

    cbpgmove(0.0, 0.0);
    cbpgdraw(0.0, 100000.0);
    cbpgsls(1);
    cbpgsci(1);

    cbpgpage();


    cbpgpage();
    cbpgsubp(2, 2);
    cbpgpanl(1, 1);

   

    vector<string> filterList;
    for(unsigned i = 0; i < table1->columns(); i++)
    {

        string name = table1->column(i)->name();
        string::size_type k = name.find("SIGNIF");
 
        if( k != string::npos )
	{
	    if(table2->hasColumn( name ))
	    {

              string filter = name.substr(0, k - 1);
              
           

      
	      filterList.push_back(filter); 

	    }
        }

    }
    string xLabel = "Log\\d10\\u(Significance 1)";
    string yLabel = "Log\\d10\\u(Significance 2)";
    title = obsId;



  
    string columnName = "SIGNIFICANCE";





    plotData1( filterList, xLabel, yLabel, title, columnName, table1, matchedIndex1, table2, matchedIndex2);




   

    /**************************************************************
    * Plot Instrumental magnitudes
    ****************************************************************/
    xLabel = "Instrumental mag 1";

    yLabel = "Instrumental mag 2";


    plotData3(filterList, xLabel, yLabel, title, "INS_MAG", table1, matchedIndex1, table2, matchedIndex2);



    /**************************************************************
    *
    * Plot corrected count-rates
    *
    ****************************************************************/
    xLabel = "Log\\d10\\u(Corrected rate 1)";

    yLabel = "Log\\d10\\u(Corrected rate 2)";





    plotData1(filterList, xLabel, yLabel, title, "CORR_RATE", table1, matchedIndex1, table2, matchedIndex2);





    plotData2(filterList, "EFLAG 1 - EFLAG 2", " ", title, table1, matchedIndex1, table2, matchedIndex2);


 


    dataSetServer->close(dataSet1); // Finished with the file
    dataSetServer->close(dataSet2); // Finished with the file

    cbpgclos();



}

/****************************************************************************
*
*
*
*
*****************************************************************************/
void plotData1(vector<string> &filterList, string xLabel, string yLabel, string title, string columnName, 
              Table *table1, vector < int >  &matchedIndex1, 
              Table *table2, vector < int >  &matchedIndex2 )
{


 
    /******************************************************************************
    *
    *
    ******************************************************************************/
    for(unsigned i = 0; i < filterList.size(); i++)
    {
        string filter = filterList[i];



	//	cout << i << ") line 2330 filter=" << filter << endl;


        vector<float> xData1, yData1;
        int size = 0;
        string name = filter + "_EXTENDED_FLAG";

   

        vector<int8> eFlag1, eFlag2;


        if ( !LoadVectorInt8ColumnData( table1, name, size, eFlag1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }


      

        if ( !LoadVectorInt8ColumnData( table2, name, size, eFlag2) )
        {
         
            Sas::fatal("Error loading column from SRCLIST table-", name);

        }
        name = filter + "_" + columnName;;






        if ( !LoadVectorFloatColumnData( table1, name, size, xData1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }

	//	cout << "Line 2374 size=" << size << endl;


        if ( !LoadVectorFloatColumnData( table2, name, size, yData1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }


        vector<float> xData, yData;
        vector<int8> eFlag;  

    

        for(unsigned i = 0; i < matchedIndex1.size(); i++)
        {
            int k1 = matchedIndex1[i];
            int k2 = matchedIndex2[i];

      


            if ( !isNull(xData1[ k1 ] ) && !isNull( yData1[ k2 ] ) &&
		 xData1[ k1 ] > 0.0 && yData1[ k2 ] > 0.0)


            {
            
	        xData.push_back( log10( xData1[ k1 ] ) );
                yData.push_back( log10( yData1[ k2 ] ) );
                eFlag.push_back( max( eFlag1[ k1 ], eFlag2[ k2 ]) );
            	

               
            }
        }
   


	if(xData.size() < 5) continue;
      
       
   
 
        float xMin = floor( MinVal( xData ) * 10.0) / 10.0;
        float xMax = ceil( MaxVal( xData )  * 10.0) / 10.0;;
 
        float yMin = floor( MinVal( yData ) * 10.0) / 10.0;
        float yMax = ceil( MaxVal( yData )  * 10.0) / 10.0;;
 
        xMin = min( xMin, yMin);
        xMax = max( xMax, yMax);
        xMax = min(float(2.0), xMax);
    
        yMin = xMin;
        yMax = xMax;

        string text = title + " " + filter + " comparison";
        drawGraph( xData, yData, 
              (char *)text.c_str(), (char *)xLabel.c_str(), (char *)yLabel.c_str(), 
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, false, false, true);


     
   
 

	vector<float> ratio;
        for(unsigned i = 0; i < xData.size(); i++)
        {
	    ratio.push_back( xData[i] / yData[i] );
        }
        plotStatistics(ratio, "x/y");



    
   
 

        cbpgsci(1);
    
        cbpgmove( -10000.0, -10000.0 );
        cbpgdraw( +10000.0, +10000.0 );

        cbpgsci(2);

      
   
 

        for(unsigned i = 0; i < xData.size(); i++)
        {

            if((int)eFlag[i] == 1)
	    {
                cbpgpt1( xData[i], yData[i], 9);


            }

        }


    
 

        cbpgsch(1.2);
     
        cbpgmtxt("T", -2.0, 0.0, 0.0, "Extension flag=1");


     
   
 

        cbpgsci( 1 );

        cbpgpage();

    } // Finished looping through the filters

       
   
 

}
/****************************************************************************
*
*
*
*
*****************************************************************************/
void plotData2(vector<string> &filterList, string xLabel, 
               string yLabel, string title, 
               Table *table1, vector < int >  &matchedIndex1, 
               Table *table2, vector < int >  &matchedIndex2 )
{



    /******************************************************************************
    *
    *
    *
    ******************************************************************************/

    for(unsigned i = 0; i < filterList.size(); i++)
    {
        string filter = filterList[i];
        vector<float> xData1, yData1;
        int size = 0;
        string name = filter + "_EXTENDED_FLAG";


 

        vector<int8> eFlag1, eFlag2;


        if ( !LoadVectorInt8ColumnData( table1, name, size, eFlag1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }



        if ( !LoadVectorInt8ColumnData( table2, name, size, eFlag2) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }


        vector<int16> eFlag;
        vector<int16> eFlagDiff;

        for(unsigned i = 0; i < matchedIndex1.size(); i++)
        {
	 

            int k1 = matchedIndex1[i];
            int k2 = matchedIndex2[i];
            if (!isNull( eFlag1[k1]) && !isNull(eFlag2[k2] ) )
	    {
	  
 	
                eFlag.push_back( max( eFlag1[ k1 ], eFlag2[ k2 ]) );
                eFlagDiff.push_back(  eFlag1[ k1 ] - (int)eFlag2[ k2 ] );
            }
           
        }
        if((int)eFlagDiff.size() < 3 ) continue;   

       string text = title + " " + filter + " comparison";
 
       if(yLabel.length() ){};



       Histogram1( eFlagDiff, (int16)1, 
	           (char *)xLabel.c_str(), (char *)text.c_str(), 
	            0.15, 0.9, 0.15, 0.7, (int16)-2, (int16)2);


       cbpgsls(2);
       cbpgsci(2);


       cbpgmove(0.0, 0.0);
       cbpgdraw(0.0, 100000.0);

       cbpgsls(1);

       cbpgsci(1);



        cbpgpage();

    } // End of looping through the filters
}










/****************************************************************************
*
*
*
*
*****************************************************************************/
void plotData3(vector<string> &filterList, string xLabel, string yLabel, string title, string columnName, 
              Table *table1, vector < int >  &matchedIndex1, 
              Table *table2, vector < int >  &matchedIndex2 )
{


  // cout << "plotdata1 line 2319\n";
    /******************************************************************************
    *
    *
    ******************************************************************************/
    int size = 0;
    string name = "RA_CORR";
    vector<double> ra1;
    if ( !LoadVectorDoubleColumnData( table1, name, size, ra1) )
    {
            Sas::fatal("Error loading column from SRCLIST table-", name);


    }
    name = "RA";


    vector<double> ra2;
    if ( !LoadVectorDoubleColumnData( table2, name, size, ra2) )
    {
            Sas::fatal("Error loading column from SRCLIST table-", name);


    }

    name = "DEC_CORR";
    vector<double> dec1;
    if ( !LoadVectorDoubleColumnData( table1, name, size, dec1) )
    {
            Sas::fatal("Error loading column from SRCLIST table-", name);


    }

   name = "DEC";

    vector<double> dec2;
    if ( !LoadVectorDoubleColumnData( table2, name, size, dec2) )
    {
            Sas::fatal("Error loading column from SRCLIST table-", name);


    }








    for(unsigned i = 0; i < filterList.size(); i++)
    {
        string filter = filterList[i];



	//	cout << i << ") line 2330 filter=" << filter << endl;


        vector<float> xData1, yData1;
        int size = 0;
        string name = filter + "_EXTENDED_FLAG";

   

        vector<int8> eFlag1, eFlag2;


        if ( !LoadVectorInt8ColumnData( table1, name, size, eFlag1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }


	//	cout << "Line 2350\n";

        if ( !LoadVectorInt8ColumnData( table2, name, size, eFlag2) )
        {
         
            Sas::fatal("Error loading column from SRCLIST table-", name);

        }
        name = filter + "_" + columnName;;




	//	cout << "Line 2363\n";



        if ( !LoadVectorFloatColumnData( table1, name, size, xData1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }

	//	cout << "Line 2374 size=" << size << endl;


        if ( !LoadVectorFloatColumnData( table2, name, size, yData1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }




        name = filter + "_QUALITY_FLAG";;




        vector<int16> qFlag1;
        if ( !LoadVectorInt16ColumnData( table1, name, size, qFlag1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }
        vector<int16> qFlag2;
        if ( !LoadVectorInt16ColumnData( table2, name, size, qFlag2) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }


        name = filter + "_CORR_RATE";;
         
	vector<float> corrRate1, corrRate2;
        if ( !LoadVectorFloatColumnData( table1, name, size, corrRate1) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }

    

        if ( !LoadVectorFloatColumnData( table2, name, size, corrRate2) )
        {
            Sas::fatal("Error loading column from SRCLIST table-", name);


        }

        vector<float> xData, yData;
        vector<int8> eFlag;  

       
        
        for(unsigned i = 0; i < matchedIndex1.size(); i++)
        {
            int k1 = matchedIndex1[i];
            int k2 = matchedIndex2[i];

            if ( !isNull(xData1[ k1 ] ) && !isNull( yData1[ k2 ] ) )
            {
            
	        if( xData1[k1] > 0.0 && yData1[k2] > 0.0)
	        {
	            xData.push_back( xData1[ k1 ]  );
                    yData.push_back( yData1[ k2 ]  );

                    eFlag.push_back( max( eFlag1[ k1 ], eFlag2[ k2 ]) );
                    
            
              
            
                }
            }
        }
       
	if(xData.size() < 5) continue;
	


   
 
        float xMin = floor( MinVal( xData ) * 10.0) / 10.0;
        float xMax = ceil( MaxVal( xData )  * 10.0) / 10.0;;
 
        float yMin = floor( MinVal( yData ) * 10.0) / 10.0;
        float yMax = ceil( MaxVal( yData )  * 10.0) / 10.0;;
 
        xMin = min( xMin, yMin);
        xMax = max( xMax, yMax);


	//        xMax = min(float(2.0), xMax);
    
        yMin = xMin;
        yMax = xMax;




        string text = title + " " + filter + " comparison";

      


        drawGraph( xData, yData, 
              (char *)text.c_str(), (char *)xLabel.c_str(), (char *)yLabel.c_str(), 
               0.15, 0.95, 0.15, 0.95, xMin, xMax, yMin, yMax, false, false, true);



	vector<float> ratio;
        for(unsigned i = 0; i < xData.size(); i++)
        {
	    ratio.push_back( xData[i] - yData[i] );
        }

        plotStatistics(ratio, "x-y");


        cbpgsci(1);
    
        cbpgmove( -10000.0, -10000.0 );
        cbpgdraw( +10000.0, +10000.0 );

        cbpgsci(2);

        for(unsigned i = 0; i < xData.size(); i++)
        {

            if((int)eFlag[i] == 1)
	    {
                cbpgpt1( xData[i], yData[i], 9);


            }

        }



        cbpgsch(1.2);
     
        cbpgmtxt("T", -2.0, 0.0, 0.0, "Extension flag=1");

        cbpgsci( 1 );

        cbpgpage();

    } // Finished looping through the filters


}








/***********************************************************
*
*  Sets all the rows in all the columns in a table to null
*  except bool and string
*
************************************************************/
void setNulls(Table *table)
{

    unsigned size = table->rows();
    for(unsigned i = 0; i < table->columns(); i++)
    {
        Column *column = table->column(i);

        switch( column -> dataType() )
        {
        case Column::Bool :
          
	    break;
	case Column::Int8 :
	{	  
   
            
            ColumnData *c = column->data();

            column -> nullValue( 128 );

	    for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;
  
 
	    break;
        }
	case Column::Int16 :
	{  
	    
            ColumnData *c = column->data();
   
            column -> nullValue( 32767 );

            for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;
 
	    break;
        }
	case Column::Int32 :
	{
	     
            column -> nullValue( 32767 );

            ColumnData *c = column->data();

            for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;
 
	    break;
        }
	case Column::Real32 :
	{
          
            ColumnData *c = column->data();

            for(unsigned i = 0; i < size; i++)
            {
    
                c -> setToNull( i );
  
            }
            delete c;  
	    break;
	}
	case Column::Real64 :
	{
             
            ColumnData *c = column->data();

            for(unsigned i = 0; i < size; i++)
            {    
                c -> setToNull( i );
            }

            delete c;
 
	    break;
          }
	case Column::DString :    // Need  to do
	{
         
	    break;
        }

	default :
	    break;
	} // switc

    }


}




/***********************************************************
*
*  Sets all the rows in all the columns in a table to null
*  except bool and string
*
************************************************************/
void setNulls(Table *table, string filter, int index)
{
  
    string filt = filter + "_"; 
    for(unsigned i = 0; i < table->columns(); i++)
    {
        Column *column = table->column(i);
        string name = table->column(i)->name();
     

        string::size_type k = name.find( filt );
      
     
    
 
        if( k == string::npos ) // 
	{
	    continue;  // Ignore this column

        }

	//      cout << "setnulls line 2686 column =" << name << " index=" << index << endl;
        switch( column -> dataType() )
        {
        case Column::Bool :
          
	    break;
	case Column::Int8 :
	{	  
   
            
            ColumnData *c = column->data();

            column -> nullValue( 128 );
    
            c -> setToNull( index );
  
            
            delete c;
  
 
	    break;
        }
	case Column::Int16 :
	{  
	    
            ColumnData *c = column->data();
   
            column -> nullValue( 32767 );

            
    
            c -> setToNull( index );
  
            delete c;
 
	    break;
        }
	case Column::Int32 :
	{
	     
            column -> nullValue( 32767 );

            ColumnData *c = column->data();
    
            c -> setToNull( index );
  

            delete c;
 
	    break;
        }
	case Column::Real32 :
	{
          
            ColumnData *c = column->data();

         
    
            c -> setToNull( index );
  
            
            delete c;  
	    break;
	}
	case Column::Real64 :
	{
             
            ColumnData *c = column->data();

               
            c -> setToNull( index );
            

            delete c;
 
	    break;
          }
	case Column::DString :    // Need  to do
	{
         
	    break;
        }

	default :
	    break;
	} // switch

    }


}

/***************************************************************
*
* Plots various statistics from data stored in x at the top
* of a graph
*
**************************************************************/

void plotStatistics(vector<float> &x, string text)
{
    sort(x.begin(), x.end());
    //    float median = Median(x);

    float mean = Mean(x);
    float p1 = Percentile(x, 0.05);
    float p2 = Percentile(x, 0.95);

    char text1[200];
    sprintf(text1, "%s mean=%4.2f, 0.05 percentile=%4.2f, 0.95=%4.2f", 
                 (char *)text.c_str(), mean, p1, p2);

    cbpgsch(1.2);
     
    cbpgmtxt("T", -1.0, 0.0, 0.0, text1);

 
}
/*****************************************************************************
*
*
*
*****************************************************************************/
void loadData(string srcFile,  vector < OmSourceParameter > &srcList, int type)
{ 


     errstr << "Loading data from file " << srcFile 
	   << Sas::message(Msg::App, Msg::Sparse);


    string filter;

    if(type == 1)
    {
        /***************************************************************
	*
        * exposure SWS file
        *
        ***************************************************************/ 
        int size = 0; 
        if ( !loadSourceParameters( srcFile, "SRCLIST", size, srcList, filter) )
        {

            errstr << "Could not load data from file  " << srcFile
	           << Sas::message(Msg::App, Msg::Sparse);

  
        }
    }
    else
    {
        /***************************************************************
	*
        * observation OBS file
        *
        ***************************************************************/
        loadObsSourceParameters( srcFile, srcList );
       
    }  

}

/******************************************************************************
*
*
*
*
*******************************************************************************/
void  refineMatchedList(const string file, vector<int> &unmatchedIndex)
{



    DataSet *dataSet = dataSetServer->open(file, Dal::Read);
    Table *table = dataSet->table("SRCLIST");
  

    vector<string> filterList;

    /********************************************************************
    *
    * Loop through all the columns in table 1
    *
    *
    *********************************************************************/
    for(unsigned i = 0; i < table->columns(); i++)
    {     
        string name = table->column(i)->name();
        string::size_type k = name.find("SIGNIF");
 
        if( k != string::npos )
	{
	    string filter = name.substr(0, k - 1);      
	    filterList.push_back(filter); 

	}

    }       // End of looping through all the columns
  
    map<string, vector<int> > indexMap;
    for(unsigned i = 0; i < filterList.size(); i++)
    {

        /**********************************************************************
        *
	* Get the RAW-RATE values for filter i
        *
	**********************************************************************/
       
  
        vector<float> rate;
        int size = 0;
        
     
	string name = filterList[i] + "_RAW_RATE";
        if ( !LoadVectorFloatColumnData( table, name, size, rate) )
	{



	}
        /***************************************************************
	*
        *
        *
        ****************************************************************/
        for(unsigned j = 0; j < unmatchedIndex.size(); j++)
	{
	    int k = unmatchedIndex[k];
	    if( !isNull(rate[k] ) )
	    {    
                if (rate[k] <= 5.0)  // Can use this data
		{
		    indexMap[ filterList[i] ].push_back( unmatchedIndex[k] );
                }
            }
        }     
    }

    dataSetServer->close(dataSet);   // Finished with the file

}
/***********************************************************
*
* Add a column "SKYIMAGE" to say what kind of image
* produced the data
*
*************************************************************/
void addSkyImageColumn(Table *table, string name, vector<string>  &text)
{



    int width = 4;
    longVec size(1);
    size[0] = width; 
          	  
          
    Column *newColumn  = table->addColumn( name, Column::DString, " ", " ", size );
      
                  
    ColumnData *columnData = newColumn->data();

 	  
       
    for(int j = 0; j < (int)table->rows(); j++)
    { 
        setStringColumnRow( columnData, j, text[j] );
    }
 	  

}





/*************************************************************************
*
*
*
*
*************************************************************************/
void setRADecColumns(Table *table, vector<int> &index)
{


  
    if( !table->hasColumn("RA_CORR") )
    {
        return;

    }

    /********************************************************
    *
    *
    ******************************************************/
    vector<double> ra;
    int size = 0;
    if ( !LoadVectorDoubleColumnData(table, "RA", size, ra) )
    {

    }

    vector<double> dec;
    size = 0;
    if ( !LoadVectorDoubleColumnData(table, "DEC", size, dec) )
    {

    }


    vector<double> raCorr;
    size = 0;
    if ( !LoadVectorDoubleColumnData(table, "RA_CORR", size, raCorr) )
    {

    }

    vector<double> decCorr;
    size = 0;
    if ( !LoadVectorDoubleColumnData(table, "DEC_CORR", size, decCorr) )
    {

    }


    for(unsigned i = 0; i < index.size(); i++)
    {
       int k = index[i];
       raCorr[k] = ra[k];
       decCorr[k] = dec[k];



    }

    Column *column = table->column("RA_CORR");
    ColumnData* columnData = column->data(); 
    setColumnData( size, columnData, raCorr );

    column = table->column("DEC_CORR");
    columnData = column->data(); 
    setColumnData( size, columnData, decCorr );





}

/*********************************************************************************
*
*
*
**********************************************************************************/
void removeColumns(Table *table)
{

    /****************************************************
    *
    * Need to delete the columns in reverse
    *
    *****************************************************/
  
    for(int i = (int)table->columns() - 1; i >= 0; i--)
    {     
        string name = table->column(i)->name();
        string::size_type k = name.find("RAW_");
      
 
        if( k != string::npos ) // Delete the column
	{
	
	  table->deleteColumn( name );



        }

    }       // End of looping through all the columns
  
}

/*************************************************************************
*
*
* Checks the total raw-rate column 
* Only keep new sources if the total raw-rate is no greater than that
* specified by the parameter "maxrawrate"
*
* NB This routine only for use with a SRCLIST table from an observation
* source-list file
*
*************************************************************************/
void  checkNewRateColumns(Table *table, vector<int> &index)
{

 
    errstr << "Checking total raw-count rates of sources in the update table"
	   << Sas::message(Msg::App, Msg::Sparse);


    errstr << "Before deleting the new table rows with no valid total raw-count rates the total number of rows=" 
           << table->rows()
	   << Sas::message(Msg::App, Msg::Sparse);

    /****************************************************
    *
    * 
    *
    *****************************************************/  
    float maxRate = realParameter( "maxrawrate");


    map<string, vector<float> > rateMap;  // Store per filter
    vector<string> filterList;            // Stores the different filters in the table

    /************************************************************
    *
    * Loop through each column and look for a RAW_RATE column
    *
    *************************************************************/

    for(int i = 0; i < (int)table->columns(); i++)
    {     
        string name = table->column(i)->name();
        string::size_type k = name.find("RAW_RATE");
      
         
        if( k != string::npos ) // Found a RAW_RATE column
	{

	 
	    /*********************************************
	    *
	    * Extract the filter
            *
            *********************************************/
	    string filter = name.substr(0, k - 1);
	   
	  
    
            filterList.push_back(filter);

            /****************************************************************
	    * Load the rates
            *
            *****************************************************************/            

            int size = 0;

            if ( !LoadVectorFloatColumnData(table, name, size, rateMap[filter]) )
            {


            }

   
	    for(unsigned j = 0; j < index.size(); j++)
	    {
               
              int k = index[j];	    
              
	      if( !isNull( rateMap[filter][k] ) )
	      { 
	      
	          if( rateMap[filter][k] > maxRate) // Don't keep this value
		  {

	      
                    /*************************************************************
		    *
                    * Set all the values in row k of column names containing the
                    * filter to -9999.9 (null value)
                    *
                    **************************************************************/
		    setNulls( table, filter, k);

              
                    rateMap[filter][k] = -9999.9;   
		  
                  }
	      }

            }

        }

    }       // End of looping through all the columns
  

      
    /**********************************************************************************
    *
    * Remove rows when no good sources
    * Do in reverse order
    *
    **********************************************************************************/
    for(int i = (int)index.size() - 1; i >= 0; i--)
    {

        
        int k = index[i]; // Row number to check 

        /******************************************************************
	*
        * Check that there is at least one filter for the row number with 
        * a non-null rate
        *
        *******************************************************************/

        int count = 0;
        for(unsigned j = 0; j < filterList.size(); j++)
        {     
	    string filter = filterList[j];
	     	         
	    if( !isNull( rateMap[filter][k] ) )
	    {
                 if ( abs(rateMap[filter][k] + 9999.9) > 0.1 &&
                      rateMap[filter][k] > 0.0               &&
                      rateMap[filter][k] <= 5.0 )
	         {
	             ++count;
            

                 }
            }
         
        }
	
  
        if(count == 0) // Delete this row
        {
	 

	    table->deleteRows(k, 1); 

        }

    }

    errstr << "After deleting the new rows with no valid total raw-count rates the total number of rows=" 
           << table->rows()
	   << Sas::message(Msg::App, Msg::Sparse);

}

/*************************************************************
*
* Need to add a column for each filter to say which 
* RATE came from a MOSAICED sky image
*
* rowFilled is a map with key equal to the column name
*
* rowFilled will be either 0, 1, or 2
*
*************************************************************/
void addSkyImageColumns(Table *table, 
			map<string, vector<int> > &rowFilled)


{

    errstr << "Adding sky-image identifier columns"
	   << Sas::message(Msg::App, Msg::Sparse);


    vector<string> skyImage(table->rows());   


    /***************************************************************
    *
    * Loop through the columns and check the SIGNIFICANCE columns
    *
    *****************************************************************/
    int size = (int)table->rows();
    for(int i = 0; i < (int)table->columns(); i++)
    {     
        string name = table->column(i)->name();
        string::size_type k = name.find("SIG");
      
      
         
        if( k != string::npos ) // Found a SIGNIFICANCE column
	{

	  //	    cout << "Checking column " << name << endl;
	
	    /*********************************************
	    *
	    * Extract the filter
            *
            *********************************************/
	    string filter = name.substr(0, k - 1);
	   
            vector<string> isSkyImage(size, "NO");
            


	    for(int j = 0; j < size; j++)
	    {

                

	        switch ( rowFilled[name][j] )    
		{
		    case 0:
		        isSkyImage[j] = "NULL";
                        break;
                    case 1:
		        isSkyImage[j] = "NO";
                        break;
                    case 2:
		      //		      cout << "Setting row " << j << " to yes\n";
		        isSkyImage[j] = "YES";
                        break;

                }

            }
            string columnName = filter + "_SKY_IMAGE";
            addSkyImageColumn(table, columnName, isSkyImage);

        }
    }



}
    

/****************************************************************
*
* Check to see if any of the matched sources in srclist2 have a 
* raw-rate no greater than the maxrawrate value but a higher 
* significance than the matched source in srclist1
*
*****************************************************************/
void checkSignificances(Table *table, Table *updateTable, 
                        vector<int> &matchedIndex1, 
                        vector<int> &matchedIndex2, 
                        map<string, vector<int> > rowFilled)
{

  

    errstr << "Comparing significances"
	   << Sas::message(Msg::App, Msg::Sparse);

    
       

    


    float maxRate = realParameter("maxrawrate");



    /***************************************************************
    *
    * Loop through the columns and check the SIGNIFICANCE columns
    *
    *****************************************************************/
    int noReplaced  = 0;   


    for(int i = 0; i < (int)table->columns(); i++)
    {

        string name = table->column(i)->name();
        string::size_type k = name.find("SIGNIF");
      
      
         
        if( k != string::npos ) // Found a SIGNIFICANCE column
	{
	
	    /*********************************************
	    *
	    * Extract the filter
            *
            *********************************************/
	    string filter = name.substr(0, k - 1);
	   
            
	    /*****************************************************************************************
	    *
            * Check to see if this filter exists in srclist2
            *
            ********************************************************************************************/  
            if (!updateTable->hasColumn( name ) ) 
            {


                continue;  // Filter doesn't exist in the update table
            }



            errstr << "Comparing significances for filter " << filter
	           << Sas::message(Msg::App, Msg::Sparse);

            vector<float> sig1;

            int size = 0;
            if ( !LoadVectorFloatColumnData( table, name, size, sig1))
            {

            }
         
            vector<float> sig2;
          
            if ( !LoadVectorFloatColumnData( updateTable, name, size, sig2) )
            {

            }
            name = filter + "_RAW_RATE"; 
            vector<float> rate2;
            if ( !LoadVectorFloatColumnData( updateTable, name, size, rate2) )
            {

            }
         
        
 
            /****************************************************************************
	    *
            * Compare matched significances
	    *
            *****************************************************************************/
            vector<int> rowNumber;
            vector<int> updateIndex;

            for(unsigned j = 0; j < matchedIndex1.size(); j++)
	    { 
	                  
	        int k = matchedIndex1[j];
		if ( rowFilled[name][k] != 1) continue;  // Ignore this row
                int l = matchedIndex2[j];
                if ( isNull( sig2[l]) ) continue;	
		if ( rate2[l] > maxRate) continue; 

                if (sig2[l] > sig1[k] ) // Replace values
		{

		    rowNumber.push_back(k);
                    updateIndex.push_back(l);
		    for(unsigned n = 0; n < updateTable->columns(); ++n)
		    { 
                         
		        string columnName = updateTable->column(n)->name();
                        string::size_type k = name.find( filter );
                        
                        if( k != string::npos ) // Found a FILTER column in the update table- replace all values
			{
			    rowFilled[columnName][k] = 0; 
                            

                        }
                        

                    }

                    /*************************************************************
                    *
                    *  Replace the data for the filter columns in the output table
                    *  with the data from the update column
                    * 
                    *  
                    *
                    **************************************************************/
                    fillTable( table, rowNumber, updateTable, updateIndex, rowFilled, 2 );

		  ++noReplaced;


                } 

            }

        }

    } // End of looping through all the columns in the output table







    //    cout << "Number of values replaced=" << noReplaced << endl;



}

/**********************************************************************
*
* Need to set the RA_CORR and DEC_CORR columns in the output table
* for those unmatched sources in the update table
* The coordinates will have been modified in the matching process
* ie- aligned with those in the first source-list file
***********************************************************************/
void setCorrectedCoordinates(Table *table,  
                             vector < OmSourceParameter > &srcList2,
                             vector<int> &rowNums, vector<int> &updateNums)
{


  // cout << "SetcorrectedCoordinates line 3777\n";

    if( !table->hasColumn("RA_CORR") ) // If no RA_CORR/DEC_CORR columns return
    {
        return;

    }



    // cout << "SetcorrectedCoordinates line 3786\n";

    Column *col1 = table -> column( "RA_CORR" );
    ColumnData *c1 = col1 -> data();
    double *ra = c1 -> real64Data();


    Column *col2 = table -> column( "DEC_CORR" );
    ColumnData *c2 = col2 -> data();

    double *dec = c2 -> real64Data();


    for (unsigned i = 0; i < rowNums.size(); i++)
    {
        int j = rowNums[i];     // The row number in the output source-list file
        int k = updateNums[i];  // The number in the second input source-list file
	// cout << i << ") j=" << j << " k=" << k << endl;  
        ra[j] =   srcList2[ k ].rightascension ;
        dec[j] =  srcList2[ k ].declination;
	//        ra[j]  = -1.0;
	// dec[j] = -1.0;

     
    }   
    delete c1;
    delete c2;
}
/*********************************************************
*
*
*
**********************************************************/
void checkTable(Table *table, string colName)
{
    if (!table->hasColumn(colName) )
    {
      cout << "No such column, " << colName << " in table " << table->name() << endl;
      return;
    }
    cout << "Checking column " << colName << endl;
    Column *col = table->column( colName); 

    ColumnData * c = col -> data();
    double *data = c->real64Data();
    for(unsigned i = 0; i < table->rows(); i++)
    {
	  cout << i << ") " << data[i] << endl; 
   

        if( isNull(data[i]) )
        {
	  cout << i << ") " << data[i] << endl; 
          exit(0);
        }
    }

    delete c;
}
/********************************************************************
*
* Want to check which file has a RAW_RATE column
* File 2 should have this column, but if file 1 has it then 
* need to swap file names
*********************************************************************/
void checkFiles(string &srcFile1, string &srcFile2)
{


    DataSet *dataSet = dataSetServer->open(srcFile2, Dal::Read);
    Table *table = dataSet->table("SRCLIST");

    for(int i = 0; i < (int)table->columns(); i++)
    {     
        string name = table->column(i)->name();
        string::size_type k = name.find("RAW_RATE");
      
         
        if( k != string::npos ) // Found a RAW_RATE column
	{
        
            dataSetServer->close(dataSet); // Finished with the output file
            return; 

        }

    }
    dataSetServer->close(dataSet); // Finished with the output file
       

    dataSet = dataSetServer->open(srcFile1, Dal::Read);
    table = dataSet->table("SRCLIST");

    for(int i = 0; i < (int)table->columns(); i++)
    {     
        string name = table->column(i)->name();
        string::size_type k = name.find("RAW_RATE");
      
         
        if( k != string::npos ) // Found a RAW_RATE column
	{
        
            dataSetServer->close(dataSet); // Finished with the output file
            string file = srcFile1;
            srcFile1 = srcFile2;
            srcFile2 = file;
            setParameter("srclist1", srcFile1);
            setParameter("srclist2", srcFile2);

	   
            return; 

        }

    }

    dataSetServer->close(dataSet); // Finished with the output file
    

    Sas::fatal("InvalidFiles", "No mosaiced observation source-list file");
  

}





