#ifndef SOURCELIST_H
#define SOURCELIST_H
#include <vector>
#include <f90cxx.h>
#include <Dal.h>
#include <errstr.h>
#include <iostream>

#include "OmSourceParameters.h"


struct OmSourceParameter;







/********************************************************
*
*
******************************************************/




void setStringColumnRow( ColumnData *d, int i,  string data); 

void setStringColumnRow( ColumnData *d, int i,  char* data); 




/********************************************************
*
*
******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                const string& columnName, 
	        int &size, vector<float> &data);

/********************************************************
*
*
******************************************************/
bool LoadVectorDoubleColumnData(const Table *table, 
                const string& columnName, 
	        int &size, vector<double> &data);




/********************************************************
*
*
******************************************************/
bool LoadVectorInt8ColumnData(const Table *table, 
                const string& columnName, 
	        int &size, vector<int8> &data);





/********************************************************
*
*
******************************************************/
bool LoadVectorInt16ColumnData(const Table *table, 
                const string& columnName, 
	        int &size, vector<int16> &data);






int GetNoRows(const string& fileName, const string& tableName);

float *LoadColumnData(const string &fileName, const 
       string &tableName, const string& columnName, int &size);


double *LoadDoubleColumnData(const string &fileName, const 
       string &tableName, const string& columnName, int &size);






/********************************************************
*
*
******************************************************/

vector<string> LoadVectorStringColumnData(const string& fileName, 
                       const string& tableName, 
                       const string& columnName, 
				       int &size);
   

 


vector<double> LoadVectorDoubleColumnData(const string& fileName, 
                      const string& tableName, 
                      const string& columnName, 
					  int &size);




vector<float> LoadVectorFloatColumnData(const string& fileName, 
                      const string& tableName, 
                      const string& columnName, 
		      int &size, int il = 0, int iu = -1);




double *LoadColumnData1(const string &fileName, const 
       string &tableName, const string& columnName, int &size);


int  *LoadIntColumnData(const string &fileName, const 
      string& tableName, const string& columnName, int &size);



vector<int> LoadVectorIntColumnData(const string &fileName, const 
      string& tableName, const string& columnName, int &size);



vector<int8> LoadVectorInt8ColumnData(const string &fileName, const 
      string& tableName, const string& columnName, int &size);










Integer1  *LoadInt1ColumnData(const string &fileName, const 
			      string& tableName, const string& columnName, int &size);








float *LoadColumnData(const Table* table, const string& columnName);


double *LoadDoubleColumnData(const Table* table, const string& columnName);


double *LoadColumnData1(const Table* table, const string& columnName);


int  *LoadIntColumnData(const Table* table, const string& columnName);











Integer1  *LoadInt1ColumnData(const Table*, const string& columnName);

Integer2  *LoadInt2ColumnData(const Table*, const string& columnName);


float GetAttribute(const string& fileName, const string& tableName, const string& name);

string GetAttribute(const string& fileName, const string& name);

bool GetBooleanAttribute(const string& fileName, const string& tableName, const string& name);

void addAttribute(const string& fileName, const string& name, const string&comment, const string & value);
void addAttribute(const string& fileName, const string& name, const string&comment, const bool & value);
void addAttribute(const string& fileName, const string& name, const string&comment, const float & value);
void addAttribute(const string& fileName, const string& name, const string&comment, const double & value);
void addAttribute(const string& fileName, const string& name, const string&comment, const int & value);


vector<float> GetAttributes(const string& fileName, const string& tableName, const string nameList[], int n);





bool deleteRowFromTable(const string& fileName, 
                        const string& tableName, 
                        const int rowNumber);

void setColumnData( const int size, const ColumnData* columnData, const real32* data, const int start = 0);
void setColumnData( const int size, const ColumnData* columnData, const real64* data, const int start = 0);
void setColumnData( const int size, const ColumnData* columnData, const int8* data, const int start = 0);
void setColumnData( const int size, const ColumnData* columnData, const int16* data, const int start = 0);
void setColumnData( const int size, const ColumnData* columnData, const int32* data, const int start = 0);
void setColumnData( const int size, const ColumnData* columnData, const int width, const char* data, const int start = 0);

// For vectors

void setColumnData( const int size, const ColumnData* columnData, const vector<real32>&  data, const int start = 0 );
void setColumnData( const int size, const ColumnData* columnData, const vector<real64>&  data, const int start = 0 );
void setColumnData( const int size, const ColumnData* columnData, const vector<int8>&    data, const int start = 0 );
void setColumnData( const int size, const ColumnData* columnData, const vector<int16>&   data, const int start = 0 );
void setColumnData( const int size, const ColumnData* columnData, const vector<int32>&   data, const int start = 0 );


/**********************************************************************************
*
**********************************************************************************/
bool appendToColumn(Table *table,
                     const char* columnName,
                     int iStart,
                     int nData,
		    vector<int8> data);

bool appendToColumn(Table *table,
                     const char* columnName,
                     int iStart,
                     int nData,
		    vector<int16> data);

bool appendToColumn(Table *table,
                     const char* columnName,
                     int iStart,
                     int nData,
		    vector<int32> data);
bool appendToColumn(Table *table,
                     const char* columnName,
                     int iStart,
                     int nData,
		    vector<real32> data);

bool appendToColumn(Table *table,
                     const char* columnName,
                     int iStart,
                     int nData,
		    vector<real64> data);


/********************************************************
*
*
******************************************************/
template <class T>
bool replaceTableColumn(const string& fileName,
                        const string& tableName,
                        const string& columnName, const T *data)
{

    if (data == 0)
    {
        cout << "Data not allocated\n";
        return false;

    }

    // Open the file
    DataSet *dataSet = dataSetServer->open(fileName, Dal::Modify);


    if (!dataSet)
    {
        cout << "failure to open data set\n";
        return false;
    }




    // Read the table

    if (!dataSet->hasBlock(tableName))
    {
      cout << "Table doesn't exist\n";
      dataSetServer->close(dataSet); // Finished with the file
      return false;

    }


    Table *table = dataSet->table(tableName);


    if (table->hasColumn(columnName))
    {

        const ColumnData *columnData = table->column(columnName)->data();
      

        int size = table->rows();

	//        setColumnData(size, columnData, data);


    
	int k = Data::Int8;
        switch( k )
        {
            case Data::Bool :

                errstr << "Boolean Null Values are not supported." << Sas::error( "notSupported" );

                break;
            case Data::Int8 :
	    {   
                int8 * i8 = columnData->int8Data();
                for(int i = 0; i < size; i++)
		  i8[i] = data[i];


                break;

            }
            case Data::Int16 :
	    {
                int16 * i16 = columnData->int16Data();
                for(int i = 0; i < size; i++)
		    i16[i] = data[i];

                break;
            }
            case Data::Int32 :
	    {
                int32 * i32 = columnData->int32Data();
                for(int i = 0; i < size; i++)
		  i32[i] = data[i];

                break;
            }
            case Data::Real32 :
	    {
                real32 * r32 = columnData->real32Data();
                for(int i = 0; i < size; i++)
		  r32[i] = data[i];

                break;
            }
            case Data::Real64 :
	    {
                real64 * r64 = columnData->real64Data();
                for(int i = 0; i < size; i++)
		  r64[i] = data[i];

                break;
            }
            case Data::DString :
                errstr << "String Null Values are not supported." << Sas::error( "notSupported" );
                break;
            default :
                errstr << "Invalid data type." << Sas::error( "internalError" );
                break;
		} 
       
    }

    dataSetServer->close(dataSet);


    return true;
};





/********************************************************
*
*
******************************************************/
bool addColumn(const string fileName,
	       const string tableName,   
               const string columnName, 
               const string units,
               const string comment,   
	       int nData,
	       vector<real32> &data);


/********************************************************
*
*
******************************************************/
bool addColumn(const string fileName,
	       const string tableName,   
               const string columnName, 
               const string units,
               const string comment,   
	       int nData,
	       vector<int16> &data);







bool addColumn(const string fileName,
               const string tableName,
               const string columnName,
               const string units,
               const string comment,
               int columnWidth, 
               int nData,
               vector<string> &data);




bool createSrcListFile(const string fileName, const string imagefile, vector< OmSourceParameter > & parameters);

bool appendToSrcListFile(const string fileName, vector< OmSourceParameter > & parameters);



bool upDateSrcListFile(const string &fileName, const char* tableName,
                       vector< OmSourceParameter > & parameters);



bool loadSourceParameters(const string &fileName, const char* tableName,
                          int& size, vector < OmSourceParameter > &parameters, string &filter);


bool loadSourceParameters(const string &fileName, const char* tableName,
                          int& size, vector < OmSourceParameter > &parameters, bool &oldData);


bool loadImage( const string & imageFile, const char *name, int &nx, int &ny, float image[]);

bool loadImage( const string & imageFile, const char *name, int &nx, int &ny, int8 image[]);





bool loadImage( const string & imageFile, const char *name, int &nx, int &ny, bool image[]);

#endif
