#ifndef UTILITY_H
#define UTILITY_H

#include "Dal.h"
#include <cmath>
#include<vector>
#include<iostream>
#include <cstdlib>
#include <cerrno>
#include <string>
#include <qmessagebox.h>
#include<numeric>
#include <sstream>


const double TWO_PI = 2.0 * M_PI;
const double radiansToDegrees = (double)180.0 / M_PI;
const double radiansToArcSecs = (double)(180 * 3600) / M_PI;



const double degreesToRadians = M_PI / (double)180.0;



double distance(double r1, double d1, double r2, double d2);

double Error(double mag1, double mag2);



vector<vector<int8> >   loadInt8Image(const DataSet *set, const char *name, int &nx, int &ny);
vector<vector<int16> >  loadInt16Image(const DataSet *set, const char *name, int &nx, int &ny);



vector<vector<int8> >   loadInt8Image(const string set, const char *name, int &nx, int &ny);
vector<vector<float> >  loadFloatImage(const string set, const char *name, int &nx, int &ny);



bool modifyInt16Image(DataSet *set, const char * name, const vector<vector<int16> > &image);



template<class kind>
void setBit(kind & value, int no)
{

    int l = sizeof(kind) * 8;
  
 
    value |= (1 << (no % l));


}
template<class kind>
void unsetBit(kind &value, int no)
{
    int l = sizeof(kind) * 8;
 
   
    value &= ~(1 << (no % l));


}
template<class kind>
bool bitSet(kind value, int no)
{

   int l = sizeof(value) * 8;
 

  return value& (1 << (no % l));

}

template<class T>
vector<T> difference(vector<T> &x, vector<T> &y, T constant=1);


template<class T>
vector<T> difference(vector<T> &x, vector<T> &y, T constant)
{
    vector<T> diff(x.size() );
    for(unsigned i = 0; i < x.size(); i++)
      diff[i] = (x[i] - y[i]) * constant;

    return diff;


}

/**************************************************************
*
*
***************************************************************/

bool isNull(int8 value);
bool isNull(int16 value);
bool isNull(int32 value);
bool isNull( float value);
bool isNull( double value);

bool isNull( char * value);




















/***************************************************************
*
*
****************************************************************/
/***************************************************************
*
*
****************************************************************/
//template<class T>
//T Mean(vector<T> &data)
//{

//    if (data.size() == 0) return (T)0.0;


//    return (T)(accumulate( data.begin(), data.end(), 0.0 ) / double(data.size() ));


//}
/***************************************************************
*
*
****************************************************************/
//template<class T>
//T Median(vector<T> &data1)
//{
//    vector<T> data = data1;
//    sort(data.begin(), data.end() );
//    T median;
//    int nData = data.size();

//    if (nData == 0) return 0.0;
//    if (nData == 1) return data[0]; 

//    int n2 = nData / 2;

//    if ( n2 * 2 == nData ) // Even number of points
//        median = (data[n2 - 1] + data[n2]) / 2.0;
//    else
//      median = data[n2];

//    return median;
//}    



 
/***************************************************************
*
*
****************************************************************/
//float Mean(float *data, long nData);

/***************************************************************
*
*
****************************************************************/
//float Mean(vector<float> &data);

/***************************************************************
*
*
****************************************************************/
//float Median(float *data, long nData);

/***************************************************************
*
*
****************************************************************/
//float Percentile(vector<float> &data, float perCent);
/***************************************************************

*
****************************************************************/
//float Percentile(float *data, long size, float perCent);





/********************************************************************
*
*********************************************************************/
//template<class kind>
//kind minVal(vector<kind> &data)
//{
//    kind minData = data[0];
//    for(unsigned i = 0; i < data.size(); i++)
//      minData = min(minData, data[i]);

//    return minData;



//}

/********************************************************************
*
*********************************************************************/
//template<class kind>
//kind maxVal(vector<kind> &data)
//{
//    kind maxData = data[0];
//    for(unsigned i = 0; i < data.size(); i++)
//      maxData = max(maxData, data[i]);

//    return maxData;



//}



/***************************************************************
*
*
****************************************************************/
//template<class kind>
//kind Percentile(vector<kind> &data1, float perCent)
//{  

//    vector<kind> data = data1;
//    sort(data.begin(), data.end() );
//    long size = data.size();
    
//    long index = (long)(perCent * (float)size + 0.00001) - 1;
//    index = max((long)0, index);
//    index = min(index, size - 1); 
     
    //   cout << "n=" << size << " percent=" << perCent << " index=" << index << " percentile=" << data[index] << endl;
    //  cout << "data[0]=" << data[0] << " data[" << (size - 1) << "]" << data[size - 1] << endl;
//    return data[index];


//}



/**************************************************************
*
**************************************************************/
template<class kind>
bool getValue(string text, kind &value, kind xMin = 0, kind xMax = 0)
{

    char *endPtr;
     
    kind value1 = (kind)strtod( text.c_str(), &endPtr);
    
    if (endPtr)
    {

        if( strlen(endPtr) > 0)
        {
     
   

            for(unsigned i = 0; i < strlen(endPtr) - 2; i++)
	    {
                if( endPtr[i] != ' ') {
                    char text1[100];
                    sprintf(text1, "Invalid number - %s - Please try again!", text.c_str());
      

                    QMessageBox::critical( 0, "OMCATALOGUE", text1);

                    return false;
                }
            }

        } 
    }

    if(errno != 0) {
    
        char text1[100];
        sprintf(text1, "Invalid number - %s (%s) - Please try again!", text.c_str(), strerror(errno));
      

        QMessageBox::critical( 0, "OMCATALOGUE", text1);
 

        return false;

    }

    if(xMin != xMax)
    {
        if(value1 < xMin || value1 > xMax) {

            char text1[100];
            sprintf(text1, "Number %e outside allowed range- %e - %e - Please try again!", (double)value1, (double)xMin, (double)xMax );
      

            QMessageBox::critical( 0, "OMCATALOGUE", text1);
 

            return false;
        }





    }
    value = (kind)value1;
 
  
    return true;

   
}

#endif
