#include "Dal.h"
#include "error.h"
#include <iostream>
#include <cstdlib>
#include<cmath>
#include "SasDalUtils.h"
#include "dal_utilities.h"
#include "DataSubspace.h"


#include<vector>





const double TWO_PI = 2.0 * M_PI;
const double radiansToDegrees = (double)180.0 / M_PI;
const double radiansToArcSecs = (double)(180 * 3600) / M_PI;
const double degreesToRadians = M_PI / (double)180.0;



/*******************************************************************
*
*******************************************************************/
double distance(double r1, double d1, double r2, double d2)
{



    return acos( sin(d1) * sin(d2) + cos(d1) * cos(d2) 
	            * cos(r1 - r2) ) * radiansToArcSecs; 


};





/******************************************************
*
*
*******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
		      int &size, vector<float> &data);

bool LoadVectorDoubleColumnData(const Table *table, 
                      const string& columnName, 
		      int &size, vector<double> &data);
void loadData(string file, vector<double> &ra, vector<double> &dec);



void loadData(string file, string colName, vector<float> &data);



/***********************************************************************
*
*
*
***********************************************************************/

template<class T>
void compareData(vector<T> & data1, vector<T> &data2, T tolerance)
{

    for(unsigned i = 0; i < min(data1.size(), data2.size()); i++)
    {
        if( abs(data1[i] - data2[i]) > tolerance)
        {
	  cout << i << ") " << data1[i] << " " << data2[i] << " diff=" << data1[i] - data2[i] << endl;

          Sas::fatal("Wrong data values in SRCLIST table-", " ");

        }


    }



}
void testDistances();
/******************************************************************
*
*
*******************************************************************/

int main()
{

    testDistances();

    vector<double> ra1, dec1;
    loadData( "test1.fits", ra1, dec1);


    vector<double> ra2, dec2; 
    loadData( "test2.fits", ra2, dec2);

    vector<double> ra3, dec3; 
    loadData( "test3.fits", ra3, dec3);

    vector<double> ra4, dec4; 
    loadData( "test4.fits", ra4, dec4);

    if(dec3.size() != dec4.size() )
    {
        cout << "number of rows in test3=" << dec3.size() << endl;
        cout << "number of rows in test4=" << dec4.size() << endl;
        cout << "The ras and decs in test3.fits\n";
        for(unsigned i = 0; i < ra3.size(); i++)
	{
	  cout << i << ") ra=" << ra3[i] << " dec=" << dec3[i] << endl; 

        } 

        cout << "The ras and decs in test4.fits\n";
        for(unsigned i = 0; i < ra4.size(); i++)
	{
	  cout << i << ") ra=" << ra4[i] << " dec=" << dec4[i] << endl; 

        } 

        cout << "test3.fits=" << ra3.size() << " test4 (ommergelists)=" << dec4.size() << endl;
        Sas::fatal("Wrong number of rows in SRCLIST table-", "test3.fits");
    }

    cout << "Checking RAs\n";

    compareData( ra3, ra4, 2.0 / 3600 );

    cout << "Checking Decs\n";

    compareData( dec3, dec4, 2.0 / 3600 );


    vector<float> sig1, sig2;
    cout << "Checkung U-RAW-RARES\n";

    loadData( "test3.fits", "U_RAW_RATE", sig1);
    loadData( "test4.fits", "U_RAW_RATE", sig2);


    float tol = 1e-6;
    compareData( sig1, sig2, tol);


    loadData( "test3.fits", "B_RAW_RATE", sig1);
    loadData( "test4.fits", "B_RAW_RATE", sig2);
    cout << "Checkung B-RAW_RATES\n";


    compareData( sig1, sig2, tol );






    return 0;
}
/***********************************************************
*
*
*
*
**********************************************************/
void loadData(string file, vector<double> &ra, vector<double> &dec)
{
    DataSet *dataSet = dataSetServer->open( file, Dal::Read);

    if (!dataSet)
    {


    }    
 
    const Table *table = dataSet->table( "SRCLIST", Dal::Read);
    if( !table) {};

    /**********************************************************************
    *
    *
    **********************************************************************/
  
    int size = 0;

 
    if ( !LoadVectorDoubleColumnData( table, "RA_CORR", size, ra))
    {

    

    }

 
    if ( !LoadVectorDoubleColumnData( table, "DEC_CORR", size, dec))
    {

    

    }

    dataSetServer->close(dataSet);

}





/***********************************************************
*
*
*
*
**********************************************************/
void loadData(string file, string colName, vector<float> &data)
{
    DataSet *dataSet = dataSetServer->open( file, Dal::Read);

    if (!dataSet)
    {


    }    
 
    const Table *table = dataSet->table( "SRCLIST", Dal::Read);
    if( !table) {};

    /**********************************************************************
    *
    *
    **********************************************************************/
  
    int size = 0;

 
    if ( !LoadVectorFloatColumnData( table, colName, size, data) )
    {


    }

    dataSetServer->close(dataSet);

}








/********************************************************
*
*
******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<float> &data)
{
     



  
  
    size = table->rows();
   
  
  
    if (table->hasColumn(columnName))
    { 
        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        float *xData = columnData->real32Data();
       


	for (int i=0; i< size; i++) 
        {
	  data.push_back( xData[i] ); 
	 
	}
     }
         
   
    return true;
};


/********************************************************
*
*
******************************************************/
bool LoadVectorDoubleColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<double> &data)
{
     



  
  
    size = table->rows();
   
  
  
    if (table->hasColumn(columnName))
    { 
        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        double *xData = columnData->real64Data();
       


	for (int i=0; i< size; i++) 
        {
	  data.push_back( xData[i] ); 
	 
	}
     }
         
   
    return true;
};

/******************************************************
*
*
*******************************************************/

void testDistances()
{
    srand ( time(NULL) );


    for(int i = 0; i < 1000; i++)
    {
        double ra1 = double( rand() % 36000) / 100.0;
        double dec1 = -80 + double( rand() % 16000) / 100.0;

        double delta = -5.0 + double( rand() % 1000) / 100.0;

	double ra2 = ra1;
        double dec2 = dec1 + delta / 3600.0;
        double d = distance(ra1 * degreesToRadians, dec1 * degreesToRadians, 
                            ra2 * degreesToRadians, dec2 * degreesToRadians);

           
        if (abs(d - abs(delta)) > 0.01)
	{

	  cout << "Computed d=" << d << " true=" << delta << " arcsecs\n";
            Sas::fatal("Wrong computed angular distance", "");
 

        }


    }
}
