#include "Dal.h"
#include <iostream>
#include "SasDalUtils.h"
#include "dal_utilities.h"
#include "DataSubspace.h"
#include<cstdlib>
#include<vector>
#include <algorithm>
#include <map>
#include<cmath>

/******************************************************
*
*
*
*
*******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
		      int &size, vector<float> &data);

bool LoadVectorDoubleColumnData(const Table *table, 
				const string& columnName, 
		      int &size, vector<double> &data);
void loadData(string file, vector<double> &ra, vector<double> &dec);


void setColumnData( const int size, const ColumnData *columnData, const vector<real32>& data );
void setColumnData( const int size, const ColumnData *columnData, const vector<real64>& data );
void setColumnData( const int size, const ColumnData *columnData, const vector<int8>& data );



const double TWO_PI = 2.0 * M_PI;
const double radiansToDegrees = (double)180.0 / M_PI;
const double radiansToArcSecs = (double)(180 * 3600) / M_PI;



const double degreesToRadians = M_PI / (double)180.0;




bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<real32> &data);

bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<real64> &data);

bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
	             vector<int8> &data);

void sortTable( string file );
void sortTable( Table **tableH );






using namespace std;
/*******************************************************
*
*
*
********************************************************/
class OMData
{
public:
    OMData() {};
    OMData(const OMData &c)
    {
        _ra = c._ra;
        _dec = c._dec;
        _significance = c._significance;
        _rawRate = c._rawRate;
        _corrRate = c._corrRate;
        _eFlag = c._eFlag;
        _filter = c._filter;
        _number = c._number;
    }
    vector<double> _ra, _dec;
    map<string, vector<float> > _significance;
    map<string, vector<float> > _rawRate, _corrRate;
    map<string, vector<int8> >  _eFlag;
    string _filter;
    int _number;
    





};

void createObservationFile1(string file, OMData &data);

void createObservationFile2(string file, OMData &data);

void createMasterFile(OMData &data, OMData &data1, OMData &data2);


class OMData1
{
public:
  OMData1() {};
  OMData1(int n, double ra): _number(n), _ra(ra) {};


  OMData1(const OMData1 &c)
  {
    _number = c._number;
    _ra = c._ra;
   





  }
  int _number;
  double _ra;
   
};
/************************************************************
*
************************************************************/
bool compare(const OMData1 &a, const OMData1 & b)
{
    return a._ra < b._ra;

};



/*******************************************************************
*
*******************************************************************/
double distance(double r1, double d1, double r2, double d2)
{



    return acos( sin(d1) * sin(d2) + cos(d1) * cos(d2) 
	            * cos(r1 - r2) ) * radiansToArcSecs; 


};

 
/**********************************************************************
*
* Create two observation source-list files test1.fits and test2.fits
* Create a merged file test3.fits from these two files. 
*
*
**********************************************************************/ 

int main()
{

    srand ( time(NULL) );

    OMData data, data1, data2;

    cout << "Creating master file test3.fits\n";

    createMasterFile(data, data1, data2);
    createObservationFile1("test3.fits", data);

    cout << "Creating output file test1.fits\n";


    createObservationFile1("test1.fits", data1);
    sortTable( "test1.fits" );


    cout << "Creating output file test2.fits\n";

    createObservationFile2("test2.fits", data2);
    sortTable( "test2.fits" );

    sortTable( "test3.fits" );

    return 0;

}
/***********************************************************
*
*
*
*
**********************************************************/
void loadData(string file, vector<double> &ra, vector<double> &dec)
{

    DataSet *dataSet = dataSetServer->open( file, Dal::Read);

    if (!dataSet)
    {


    }    
 
    const Table *table = dataSet->table( "SRCLIST", Dal::Read);
    if( !table) {};

    /**********************************************************************
    *
    *
    **********************************************************************/
  
    int size = 0;

 
    if ( !LoadVectorDoubleColumnData( table, "RA_CORR", size, ra))
    {

    

    }

 
    if ( !LoadVectorDoubleColumnData( table, "DEC_CORR", size, dec))
    {

    

    }

}
/********************************************************
*
*
******************************************************/
bool LoadVectorFloatColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<float> &data)
{
     
    size = table->rows();
   
    if (table->hasColumn(columnName))
    { 

        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        float *xData = columnData->real32Data();
       

	for (int i=0; i< size; i++) 
        {
	    data.push_back( xData[i] ); 
	 
	}
     }
         
   
    return true;

};

/********************************************************
*
*
******************************************************/
bool LoadVectorDoubleColumnData(const Table *table, 
                      const string& columnName, 
                      int &size, vector<double> &data)
{
       
    size = table->rows();
   
    if (table->hasColumn(columnName))
    { 
        const ColumnData *columnData = table->column(columnName)->data();
        if (!columnData) exit(0);
	

      
        double *xData = columnData->real64Data();
       


	for (int i=0; i< size; i++) 
        {
	  data.push_back( xData[i] ); 
	 
	}
     }
         
   
    return true;

};



/********************************************************
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<real32> &data)
{
     

  
    // Get the size of the table

    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Real32, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);


    delete columnData;


          
    return true;
};



/********************************************************
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<real64> &data)
{
     

  
    // Get the size of the table
    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Real64, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);


    delete columnData;


          
    return true;
};

/********************************************************
*
*
******************************************************/
bool addColumn(Table *table, 
                     const char* columnName, 
                     const char* units,
                     const char* comment,   
		     int nData,
                     vector<int8> &data)
{
     

  
    // Get the size of the table
    vector<unsigned long> size(1);
    size[0] = 1;
    size[1] = nData;
  
   
    // Create the column
    Column *column = table->addColumn( columnName, Dal::Int8, comment, units , size);
 
    ColumnData *columnData = column->data();

    setColumnData( ( int)size[1], columnData, data);


    delete columnData;


          
    return true;
};

/******************************************************
*
******************************************************/
void setColumnData( const int size, const ColumnData *columnData, const vector<int8>& data)
{
    int8 * r64 = columnData->int8Data();
    for(int i = 0; i < size; i++)
        r64[ i] = data[i];


}


/******************************************************
*
******************************************************/


void setColumnData( const int size, const ColumnData *columnData, const vector<real32>& data )
{
    real32 * r32 = columnData->real32Data();
    for(int i = 0; i < size; i++)
        r32[i] = data[i];

}

/******************************************************
*
******************************************************/
void setColumnData( const int size, const ColumnData *columnData, const vector<real64>& data)
{
    real64 * r64 = columnData->real64Data();
    for(int i = 0; i < size; i++)
        r64[ i] = data[i];


}





/*****************************************************************************
*
*
*
*
*
*
******************************************************************************/
void createObservationFile1(string file, OMData &data)
{

    DataSet *dataSet = dataSetServer->open(file, Dal::Create);

    *dataSet->addAttribute( "FILTER", "" ) = "n/a";

    *dataSet->addAttribute( "OBS_ID", "" ) = "666";

    if (!dataSet) return ;    


    cout << "Creating file " << file << " table size=" << data._ra.size() << endl;


    Table *table = dataSet->addTable( "SRCLIST", data._ra.size() );





    if (!addColumn( table, "RA", "degrees", "", 
		     (int)data._ra.size(), data._ra) )
    {
     
    }



    if (!addColumn( table, "DEC", "degrees", "",
		     (int)data._dec.size(), data._dec) )
    {
     
    }







    if (!addColumn( table, "RA_CORR", "degrees", "", 
		     (int)data._ra.size(), data._ra) )
    {
     
    }



    if (!addColumn( table, "DEC_CORR", "degrees", "",
		     (int)data._dec.size(), data._dec) )
    {
     
    }


    map<string, vector<float> >::iterator iter;
    for(iter = data._rawRate.begin(); iter != data._rawRate.end(); ++iter)
    { 

        string filter = iter->first;
        cout << "createobservation file line 499 filter=" << filter << endl;
        string name = filter + "_SIGNIFICANCE";


	if (!addColumn( table, (const char *)name.c_str(), "", "",
    	     (int)data._ra.size(), data._significance[filter]) )
	{
     
       }

        name = filter + "_CORR_RATE";
        if (!addColumn( table, (const char*)name.c_str(), "", "",
		     (int)data._ra.size(), data._corrRate[filter]) )
        {
     
        }

        name = filter + "_EXTENDED_FLAG";
 

        if (!addColumn( table, (const char*)name.c_str(), "degrees", "",
		     (int)data._ra.size(), data._eFlag[filter]) )
        {
     
        }

    }

    if(table){};
    dataSetServer->close( dataSet );

 
}







/*****************************************************************************
*
*
*
*
*
*
******************************************************************************/
void createObservationFile2(string file, OMData &data)
{

    DataSet *dataSet = dataSetServer->open(file, Dal::Create);

    *dataSet->addAttribute( "FILTER", "" ) = "n/a";

    *dataSet->addAttribute( "OBS_ID", "" ) = "666";

    if (!dataSet) return ;    


    cout << "Creating file " << file << " table size=" << data._ra.size() << endl;


    Table *table = dataSet->addTable( "SRCLIST", data._ra.size() );

    if (!addColumn( table, "RA", "degrees", "", 
		     (int)data._ra.size(), data._ra) )
    {
     
    }



    if (!addColumn( table, "DEC", "degrees", "",
		     (int)data._dec.size(), data._dec) )
    {
     
    }


    map<string, vector<float> >::iterator iter;
    for(iter = data._rawRate.begin(); iter != data._rawRate.end(); ++iter)
    { 

        string filter = iter->first;



       string name = filter + "_SIGNIFICANCE";


	if (!addColumn( table, (const char *)name.c_str(), "", "",
    	     (int)data._ra.size(), data._significance[filter]) )
	{
     
       }




        cout << "createobservation file line 499 filter=" << filter << endl;
        name = filter + "_RAW_RATE";

        if (!addColumn( table, (const char *)name.c_str(), "", "",
    	     (int)data._ra.size(), data._rawRate[filter]) )
        {
     
        }

        name = filter + "_CORR_RATE";
        if (!addColumn( table, (const char*)name.c_str(), "", "",
		     (int)data._ra.size(), data._corrRate[filter]) )
        {
     
        }

        name = filter + "_EXTENDED_FLAG";
 

        if (!addColumn( table, (const char*)name.c_str(), "degrees", "",
		     (int)data._ra.size(), data._eFlag[filter]) )
        {
     
        }

    }

    if(table){};
    dataSetServer->close( dataSet );

 
}



/*****************************************************************
*
*
* Creates a master merged observation source-list file composed
* of data taken form two individual observation source-list files
*
*
*****************************************************************/
void createMasterFile(OMData &data3, OMData &data1, OMData &data2)
{


    /*************************************************
    *
    * The total number of sources in file 1
    *
    *************************************************/
    const int maxSize = 1000;
    int size1 = max(10, rand() % maxSize);
    //   size1 = maxSize;
    double ra0 = rand() % 300;
    double dec0 = -20.0 + rand() % 140;
    double da = 10.0 / 3600.0;
    double ddec = 10.0 / 3600.0;

    double ra = 0.0, dec = 0.0;

    /**************************************************
    *
    * The observation file will contain two filters U and B
    *
    **************************************************/
    string filter1 = "U", filter2 = "B";

    data1._filter = filter1;  // Source-list 1 filter
    data2._filter = filter2;  // Source-list 2 filter

    /*************************************************
    *
    * The data for file 1
    * data1 refers to the data for input source-list file 1
    * data2 refers to the data for input source-list file 2
    *
    *************************************************/
    
    for(int i = 0; i < size1; i++)
    {
        double da1 = ( -999.0 + double( rand() % 2000) ) / 3600000.0;
        double ddec1 = ( -999.0 + double( rand() % 2000) ) / 3600000.0;

	da1 = 0.0, ddec1=0;
	ra = ra0 + (double)i * da + da1;
        dec = dec0 + (double)i * ddec + ddec1;

        data1._ra.push_back( ra );
        data1._dec.push_back( dec );
        float sig = 6.0 + rand() % 1000;

        data1._significance[filter1].push_back( sig );
    
        float rate = sig;

        data1._rawRate[filter1].push_back( rate );
        data1._corrRate[filter1].push_back( rate );
        data1._eFlag[filter1].push_back( (int8)1 );
 


        data3._ra.push_back( ra );
        data3._dec.push_back( dec );


        data3._significance[filter1].push_back( sig);
        data3._rawRate[filter1].push_back( rate );
        data3._corrRate[filter1].push_back( rate );
        data3._eFlag[filter1].push_back( (int8)1 );
 



        /*************************************************
	*
        * Add some data to the second source-list
        *
	***************************************************/


        if( rand() % 100 < 10)
        {


            data2._ra.push_back( ra );
            data2._dec.push_back( dec );
       
            data2._significance[filter2].push_back( sig );
            data2._rawRate[filter2].push_back( rate );
            data2._corrRate[filter2].push_back( rate );
            data2._eFlag[filter2].push_back( (int8)1 );
                 



            if( sig <= 5)
	    {
                data3._significance[filter2].push_back( sig );
                data3._rawRate[filter2].push_back( rate );
                data3._corrRate[filter2].push_back( rate );
                data3._eFlag[filter2].push_back( (int8)1);
	    }
            else
	    {
                data3._significance[filter2].push_back( -9999.9 );
                data3._rawRate[filter2].push_back( -9999.9 );
                data3._corrRate[filter2].push_back( -9999.9 );
                data3._eFlag[filter2].push_back( 127 );
            }

        }
        else
	{
            data3._significance[filter2].push_back( -9999.9 );
            data3._rawRate[filter2].push_back( -9999.9 );
            data3._corrRate[filter2].push_back( -9999.9 );
            data3._eFlag[filter2].push_back( 127 );
        }





       
    }


    /**********************************************************
    *
    * Give the second source-list a small offset in RA and DEC
    *
    ***********************************************************/

    ra0 = data1._ra[ size1 - 1 ];
    dec0 = data1._dec[ size1 - 1 ];
 

    int size2 = max(100, rand() % maxSize);
    size2 = size1;
    /*********************************************************************
    *
    * The data for file 2
    * The
    * 
    *********************************************************************/
    

    data3._filter = "n/a";


    /********************************************************************
    *
    *
    *
    *********************************************************************/   
    for(int i = 0; i < size2; i++)
    {
      
        double da1 = ( -999.0 + double(rand() % 2000)) / 3600.0;
        double ddec1 = ( -999.0 + double(rand() % 2000)) / 3600.0;
	da1 = 0.0, ddec1=0;
        ra = ra0 + (double)i * da + da1;
        dec = dec0 + (double)i * ddec + ddec1;
       
	double ra1  = ra * degreesToRadians;
	double dec1 = dec * degreesToRadians;

        bool found = false;
        for(unsigned j = 0; j < data2._ra.size(); j++)
	{

	    double d = distance(ra1, dec1, data2._ra[j]  * degreesToRadians, 
				data2._dec[j] * degreesToRadians);
        



            if(d <= 1.0)  // This source already exists
	    {
	       

                 
	        found = true;
                break;
            }

        } 
        if(found) continue;

        for(unsigned j = 0; j < data1._ra.size(); j++)
	{

	    double d = distance(ra1, dec1, data1._ra[j]  * degreesToRadians, 
				data1._dec[j] * degreesToRadians);
            if(d > 1.99 && d < 2.01) continue; 



            if(d <= 2.0)  // This source already exists
	    {
	       

                 
	        found = true;
                break;
            }

        } 
	
        data2._ra.push_back( ra );
        data2._dec.push_back( dec );
        float sig = 2.0 + rand() % 100;
        data2._significance[filter2].push_back( sig );
    
        float rate = sig;

        data2._rawRate[filter2].push_back( rate );
        data2._corrRate[filter2].push_back( rate );
        data2._eFlag[filter2].push_back( (int8)1 );
   
           
	if(found) continue;  // Already in master list

        if(sig <= 5.0)
	{
            data3._ra.push_back( ra );
            data3._dec.push_back( dec );
   
            /*****************************************************
	    *
            * 
            *
            ******************************************************/

            data3._significance[filter1].push_back( -9999.9 );
            data3._rawRate[filter1].push_back( -9999.9 );
            data3._corrRate[filter1].push_back( -9999.9 );
            data3._eFlag[filter1].push_back( 127 );
            
            data3._significance[filter2].push_back( sig );
            data3._rawRate[filter2].push_back( rate );
            data3._corrRate[filter2].push_back( rate );
            data3._eFlag[filter2].push_back( (int8)1 );

        }     
    }



    cout << "The ras and decs in test3.fits\n";
    for(unsigned i = 1; i < data3._ra.size(); i++)
    {
        cout << i << ") ra=" << data3._ra[i] << " dec=" << data3._dec[i] << endl;
        double ra1 = data3._ra[i - 1] * degreesToRadians;
        double dec1 = data3._dec[i - 1] * degreesToRadians;


        double ra2  = data3._ra[i] * degreesToRadians;
        double dec2 = data3._dec[i] * degreesToRadians;

        double d = distance(ra1, dec1, ra2, dec2);
        if(d < 1.0)
	{
	    cout << "Bug in test program d=" << d << " arcsecs\n";
            exit(0);

        } 

    }


}
/*************************************************************************
*
*
**************************************************************************/
void sortTable( string file )
{


    DataSet *dataSet = dataSetServer->open(file, Dal::Modify);
    if (!dataSet) return ;    


    

    Table *tableH = dataSet->table("SRCLIST");
  
    sortTable( &tableH );


    dataSetServer->close(dataSet); // Finished with the file





}
/*******************************************************************
*
*
*
********************************************************************/
void sortTable( Table **tableH )
{
 
    int nRows = (*tableH)->rows();


    if(nRows < 2)
    {
        return;
    }


    /**********************************************************
    * 
    *
    *********************************************************/
    vector<double> ra;

    if( (*tableH)->hasColumn("RA_CORR") ) 
    {
        int size = 0; 
        if ( !LoadVectorDoubleColumnData( *tableH, "RA_CORR", size, ra))
	{
        }
    }
    else
    {
        int size = 0;
        if ( !LoadVectorDoubleColumnData( *tableH, "RA", size, ra) )
	{
        }


    }

    vector<OMData1> data;

    for(int i = 0; i < nRows; i++)
    {

        
        data.push_back( OMData1(i, ra[i]) );
       
  

    }

    vector<int> raSortIndex(nRows);   

 
    /*****************************************************
    * Find the order of increasing  RA- store in RASortIndex
    * for adding data to source list table in  
    * increasing RA order (ra is not sorted)
    *****************************************************/
    sort( data.begin(), data.end(), compare);

    for(int i = 0; i < nRows; i++)
    {
 
        raSortIndex[i] = data[i]._number; 
    }

    //    cout << "Sorttable line the sorted data 74\n";

    //  for(int i = 0; i < nRows; i++)
    //   cout << i << ") ra=" << sourceList[i].rightascension << " index=" << raSortIndex[i] << endl;



    (*tableH)->insertRows( nRows,  nRows );
    (*tableH)->copyRows( 0, nRows, nRows );

    /*************************************************************
    * Reorder the table into increasing RA order
    **************************************************************/
  

    for(int i = 0; i <  nRows; i++)
    {
        int k = raSortIndex[i] + nRows;       
	//	cout << i << ") k=" << k << " rasortindex=" << raSortIndex[i] << endl;
        (*tableH)->copyRows(k, i, 1);
    }


    /********************************************************
    *
    * Delete the appended temporary rows
    *
    *********************************************************/


    (*tableH)->deleteRows( nRows, nRows);
 
    



 }
